# SPDX-License-Identifier: MIT
"""CSV-backed Competitor and Category Data"""

import logging
import os
import csv
from unicodedata import normalize
import grapheme

import metarace
from metarace import strops

_log = logging.getLogger('riderdb')
_log.setLevel(logging.DEBUG)

# default values when columns omitted
_RIDER_DEFAULTS = {}

# Internal rider column keys
_RIDER_COLUMNS = {
    'no': 'Rider No',
    'series': 'Series',
    'first': 'First Name',
    'last': 'Last Name',
    'org': 'Organisation',
    'cat': 'Categories',
    'nat': 'Nationality',
    'ref': 'Transponder',
    'uci': 'UCI ID',
    'dob': 'DoB',
    'sex': 'Sex',
    'note': 'Notes',
    'seed': 'Seeding',
    'data': 'Data Reference',
}

# Column strings lookup
_ALT_COLUMNS = {
    'id': 'no',
    'ride': 'no',
    'numb': 'no',
    'cid': 'no',
    'code': 'no',
    'bib': 'no',
    'inde': 'no',
    'no.': 'no',
    'seri': 'series',
    'type': 'series',
    'firs': 'first',
    'name': 'first',
    'titl': 'first',
    'last': 'last',
    'subt': 'last',
    'surn': 'last',
    'orga': 'org',
    'club': 'org',
    'team': 'org',
    'laps': 'org',
    'lap': 'org',
    'cats': 'cat',
    'cate': 'cat',
    'targ': 'cat',
    'lice': 'uci',
    'star': 'uci',
    'ucii': 'uci',
    'refi': 'ref',
    'rfid': 'ref',
    'tran': 'ref',
    'disp': 'ref',
    'gend': 'sex',
    'nati': 'nat',
    'note': 'note',
    'foot': 'note',
    'memb': 'note',
    'rank': 'seed',
    'data': 'data',
    'date': 'dob',
}

# Category columns
_CATEGORY_COLUMNS = {
    'no': 'ID',
    'series': 'Series',
    'first': 'Title',
    'last': 'Subtitle',
    'note': 'Footer',
    'org': 'Lap Prefix',
    'ref': 'Distance',
    'cat': 'Target Laps',
    'uci': 'Start Offset',
    'nat': 'Nationality',
    'dob': 'DoB',
    'sex': 'Sex',
    'seed': 'Seeding',
    'data': 'Data Reference',
}

_TEAM_COLUMNS = {
    'no': 'Code',
    'series': 'Series',
    'first': 'Name',
    'last': 'Short Name',
    'org': 'Organisation',
    'cat': 'Categories',
    'nat': 'Nationality',
    'ref': 'Start Time',
    'uci': 'UCI ID',
    'dob': 'DoB',
    'sex': 'Division',
    'note': 'Members',
    'seed': 'Seeding',
    'data': 'Data Reference',
}

# Config schema for a rider
_RIDER_SCHEMA = {
    'rtype': {
        'prompt': 'Rider',
        'control': 'section',
    },
    'no': {
        'prompt': 'Rider No:',
        'control': 'short',
        'attr': 'no',
        'defer': True,
        'default': '',
    },
    'series': {
        'prompt': 'Series:',
        'control': 'short',
        'attr': 'series',
        'defer': True,
        'default': '',
    },
    'first': {
        'prompt': 'First Name:',
        'attr': 'first',
        'defer': True,
        'default': '',
    },
    'last': {
        'prompt': 'Last Name:',
        'attr': 'last',
        'defer': True,
        'default': '',
    },
    'org': {
        'prompt': 'Organisation:',
        'attr': 'org',
        'defer': True,
        'hint': 'Club or team affiliation',
        'default': '',
    },
    'cat': {
        'prompt': 'Categories:',
        'attr': 'cat',
        'defer': True,
        'hint': 'Space separated list of categories',
        'default': '',
    },
    'nat': {
        'prompt': 'Nationality:',
        'attr': 'nat',
        'defer': True,
        'hint': '3 letter IOC country code eg: AUS, GBR, JPN',
        'control': 'short',
        'default': '',
    },
    'ref': {
        'prompt': 'Transponder:',
        'attr': 'ref',
        'defer': True,
        'control': 'short',
        'default': '',
    },
    'uci': {
        'prompt': 'UCI ID:',
        'attr': 'uci',
        'defer': True,
        'control': 'short',
        'hint': '11 digit UCI ID',
        'default': '',
    },
    'dob': {
        'prompt': 'Date of Birth:',
        'attr': 'dob',
        'control': 'short',
        'defer': True,
        'subtext': '(YYYY-MM-DD)',
        'hint': 'ISO8601 Date of birth eg: 2012-01-25',
        'default': '',
    },
    'sex': {
        'prompt': 'Sex:',
        'control': 'short',
        'subtext': '(Male|Female)',
        'defer': True,
        'hint': 'Sex of participant',
        'attr': 'sex',
        'default': '',
    },
    'seed': {
        'prompt': 'Seeding:',
        'control': 'short',
        'defer': True,
        'hint': 'Seeding, start time or ranking of rider',
        'attr': 'seed',
        'default': '',
    },
    'data': {
        'prompt': 'Data Ref:',
        'attr': 'data',
        'defer': True,
        'hint': 'Data source, or supplemental ID',
        'default': '',
    },
    'note': {
        'prompt': 'Notes:',
        'attr': 'note',
        'defer': True,
        'hint': 'Supplementary rider notes',
        'default': '',
    },
}

# Config schema for rider series
_SERIES_SCHEMA = {
    'rtype': {
        'prompt': 'Series',
        'control': 'section',
    },
    'no': {
        'prompt': 'ID:',
        'control': 'short',
        'attr': 'no',
        'hint': 'Series ID',
        'defer': True,
        'default': '',
    },
    'series': {
        'prompt': 'Series:',
        'control': 'short',
        'readonly': True,
        'attr': 'series',
        'defer': True,
        'default': '',
    },
    'first': {
        'prompt': 'Title:',
        'attr': 'first',
        'hint': 'Series title',
        'defer': True,
        'default': '',
    },
    'last': {
        'prompt': 'Subtitle:',
        'attr': 'last',
        'hint': 'Series subtitle',
        'defer': True,
        'default': '',
    },
    'note': {
        'prompt': 'Footer:',
        'attr': 'note',
        'defer': True,
        'hint': 'Supplementary footer text for reports',
        'default': '',
    }
}

# Config schema for a category
_CATEGORY_SCHEMA = {
    'rtype': {
        'prompt': 'Category',
        'control': 'section',
    },
    'no': {
        'prompt': 'ID:',
        'control': 'short',
        'attr': 'no',
        'hint': 'Category ID or handicap group',
        'defer': True,
        'default': '',
    },
    'series': {
        'prompt': 'Series:',
        'control': 'short',
        'readonly': True,
        'attr': 'series',
        'defer': True,
        'default': '',
    },
    'first': {
        'prompt': 'Title:',
        'attr': 'first',
        'hint': 'Category title',
        'defer': True,
        'default': '',
    },
    'last': {
        'prompt': 'Subtitle:',
        'attr': 'last',
        'hint': 'Category subtitle',
        'defer': True,
        'default': '',
    },
    'note': {
        'prompt': 'Footer:',
        'attr': 'note',
        'defer': True,
        'hint': 'Supplementary footer text for reports',
        'default': '',
    },
    'uci': {
        'prompt': 'Start Offset:',
        'attr': 'uci',
        'defer': True,
        'control': 'short',
        'hint': 'Start time offset from event start',
        'default': '',
    },
    'cat': {
        'prompt': 'Target:',
        'attr': 'cat',
        'control': 'short',
        'subtext': 'laps',
        'defer': True,
        'hint': 'Target number of laps for this category',
        'default': '',
    },
    'ref': {
        'prompt': 'Distance:',
        'attr': 'ref',
        'defer': True,
        'control': 'short',
        'subtext': 'km',
        'hint': 'Category distance override',
        'default': '',
    },
    'org': {
        'prompt': 'Lap Prefix:',
        'attr': 'org',
        'control': 'short',
        'defer': True,
        'hint': 'Optional category lap prefix',
        'default': '',
    },
}

# Config schema for a series
_CATEGORY_SERIES = {
    'rtype': {
        'prompt': 'Series',
        'control': 'section',
    },
    'no': {
        'prompt': 'ID:',
        'control': 'short',
        'attr': 'no',
        'hint': 'Series ID',
        'defer': True,
        'default': '',
    },
    'series': {
        'prompt': 'Series:',
        'control': 'short',
        'readonly': True,
        'attr': 'series',
        'defer': True,
        'default': '',
    },
    'first': {
        'prompt': 'Title:',
        'attr': 'first',
        'hint': 'Series title',
        'defer': True,
        'default': '',
    },
    'last': {
        'prompt': 'Subtitle:',
        'attr': 'last',
        'hint': 'Series subtitle',
        'defer': True,
        'default': '',
    },
    'note': {
        'prompt': 'Footer:',
        'attr': 'note',
        'defer': True,
        'hint': 'Supplementary footer text for reports',
        'default': '',
    },
}

# Config schema for a team
_TEAM_SCHEMA = {
    'rtype': {
        'prompt': 'Team',
        'control': 'section',
    },
    'no': {
        'prompt': 'Team Code:',
        'control': 'short',
        'attr': 'no',
        'defer': True,
        'default': '',
    },
    'series': {
        'prompt': 'Series:',
        'control': 'short',
        'attr': 'series',
        'defer': True,
        'default': '',
    },
    'first': {
        'prompt': 'Name:',
        'attr': 'first',
        'hint': 'Team name',
        'defer': True,
        'default': '',
    },
    'last': {
        'prompt': 'Short Name:',
        'attr': 'last',
        'control': 'short',
        'subtext': '(~12 characters)',
        'hint': 'Abbreviated team name for reports',
        'defer': True,
        'default': '',
    },
    'org': {
        'prompt': 'Organisation:',
        'attr': 'org',
        'defer': True,
        'hint': 'Club or team affiliation',
        'default': '',
    },
    'nat': {
        'prompt': 'Nationality:',
        'attr': 'nat',
        'defer': True,
        'hint': '3 letter IOC country code eg: AUS, GBR, JPN',
        'control': 'short',
        'default': '',
    },
    'uci': {
        'prompt': 'UCI ID:',
        'attr': 'uci',
        'defer': True,
        'control': 'short',
        'hint': 'Team UCI ID',
        'default': '',
    },
    'sex': {
        'prompt': 'Division:',
        'control': 'short',
        'subtext': '',
        'defer': True,
        'hint': 'UCI division of team eg: WTT WTW PRT',
        'attr': 'sex',
        'default': '',
    },
    'ref': {
        'prompt': 'Start Time:',
        'attr': 'ref',
        'defer': True,
        'control': 'short',
        'subtext': '(Team TT)',
        'hint': 'Team TT Start Time',
        'default': '',
    },
    'note': {
        'prompt': 'Members:',
        'attr': 'note',
        'defer': True,
        'hint': 'List of member riders in BIB.series format',
        'default': '',
    },
}

# reserved series
_RESERVED_SERIES = ('spare', 'cat', 'team', 'ds', 'series')

# legacy csv file ordering
_DEFAULT_COLUMN_ORDER = ('no', 'first', 'last', 'org', 'cat', 'series', 'ref',
                         'uci', 'dob', 'nat', 'sex', 'note', 'seed', 'data')


def primary_cat(catstr=''):
    """Return the primary cat from a catlist (legacy support)."""
    ret = ''
    cv = catstr.split()
    if cv:
        ret = cv[0].upper()
    return ret


def colkey(colstr=''):
    """Convert a column header string to a colkey."""
    col = colstr[0:4].strip().lower()
    if col in _ALT_COLUMNS:
        col = _ALT_COLUMNS[col]
    return col


def get_header(cols=_DEFAULT_COLUMN_ORDER, hdrs=_RIDER_COLUMNS):
    """Return a row of header strings for the provided cols."""
    return (hdrs[colkey(c)] for c in cols)


def cellnorm(unistr):
    """Normalise supplied string, then return only printing chars."""
    return normalize('NFC', unistr.strip()).translate(strops.PRINT_UTRANS)


class rider():
    """Rider handle."""

    def copy(self):
        """Return copy of this rider"""
        nr = rider()
        for k, v in self.__store.items():
            nr.__store[k] = v
        nr.__notify = self.__notify
        return nr

    def get_id(self):
        """Return this rider's unique id"""
        return (self.__store['no'].upper(), self.__store['series'].lower())

    def get_schema(self):
        """Return a schema for this rider object"""
        ret = _RIDER_SCHEMA
        if self.__store['series'] == 'cat':
            ret = _CATEGORY_SCHEMA
        elif self.__store['series'].startswith('t'):
            ret = _TEAM_SCHEMA
        elif self.__store['series'] == 'series':
            ret = _SERIES_SCHEMA
        return ret

    def rename(self, listname, notify=True):
        """Replace first, last and organisation from provided listname

        Rider: 			Firstname LAST NAME (Organisation)
        Team/Cat/Series:	First Name (Organisation)

        """
        first = []
        last = []
        org = []
        inorg = False
        label = self.get_label()
        dolast = not (self.__store['series'].startswith('t')
                      or self.__store['series'] in ('series', 'cat'))
        for namebit in listname.split():
            if namebit.startswith('('):
                inorg = True
            if inorg:
                org.append(namebit.replace('(', '').replace(')', ''))
                if namebit.endswith(')'):
                    inorg = False
            elif first and dolast:
                last.append(namebit)
            else:
                first.append(namebit)
        donotify = False
        newfirst = ' '.join(first)
        if newfirst != self.get_value('firsname'):
            self.set_value('firsname', newfirst)
            donotify = True
        neworg = ' '.join(org)
        if neworg != self.get_value('organisation'):
            self.set_value('organisation', neworg)
            donotify = True
        if dolast:
            newlast = ' '.join(last)
            if newlast != self.get_value('lastname'):
                self.set_value('lastname', newlast)
                donotify = True
        if notify and donotify:
            self.__notify(self.get_id())

    def get_label(self):
        """Return a type label for this object"""
        ret = 'Rider'
        series = self.__store['series'].lower()
        if series == 'cat':
            ret = 'Category'
        elif series.startswith('t'):
            ret = 'Team'
        elif series == 'series':
            ret = 'Number Series'
        elif series == 'ds':
            ret = 'DS'
        elif series == 'spare':
            ret = 'Spare Bike'
        return ret

    def get_bibstr(self):
        """Return the bib.series string"""
        ret = None
        nkey = 'bs'
        if nkey not in self.__strcache:
            ret = strops.bibser2bibstr(self.__store['no'],
                                       self.__store['series'])
            self.__strcache[nkey] = ret
        else:
            ret = self.__strcache[nkey]
        return ret

    def get_key(self):
        """Return a sorting key for this rider number"""
        return strops.bibstr_key(self.__store['no'])

    def primary_cat(self):
        """Return rider's primary category"""
        ret = ''
        cv = self['cat'].split()
        if cv:
            ret = cv[0].upper()
        return ret

    def set_primary_cat(self, cat):
        """Set the rider's primary category"""
        new_primary = cat.upper()
        if new_primary:
            cmap = {new_primary: True}
            old_primary = None
            member = False
            for c in self['cat'].split():
                rcat = c.upper()
                if old_primary is None:
                    old_primary = rcat
                if c == new_primary:
                    member = True
                cmap[rcat] = True
            if not member:
                # replace old_primary, rider was not already a member
                if old_primary is not None and old_primary in cmap:
                    del cmap[old_primary]
            self['cat'] = ' '.join(cmap)
        else:
            _log.warning('Cleared %s %s categories', self.get_label(),
                         self.get_bibstr())
            self['cat'] = ''

    def in_series(self, series=None):
        """Return True if rider is in the nominated number series

        If series is None, return True if the rider entry is in a
        non-reserved series
        """
        if series is None and self['series'] not in _RESERVED_SERIES:
            return True
        elif self['series'] == series:
            return True
        return False

    def in_cat(self, cat):
        """Return True if rider is in the nominated category"""
        return cat.upper() in self['cat'].upper().split()

    def get_cats(self):
        """Return a list of categories for this rider"""
        return (c.upper() for c in self['cat'].split())

    def add_cat(self, cat):
        """Append cat to rider"""
        cmap = {}
        for c in self['cat'].split():
            cmap[c.upper()] = True
        cmap[cat.upper()] = True
        self['cat'] = ' '.join((c for c in cmap))

    def del_cat(self, cat):
        """Remove cat from rider"""
        cmap = {}
        for c in self['cat'].split():
            cmap[c.upper()] = True
        rem = cat.upper()
        if rem in cmap:
            del cmap[rem]
            self['cat'] = ' '.join(cmap)

    def get_row(self, coldump=_DEFAULT_COLUMN_ORDER):
        """Return a row ready to export."""
        return (str(self[c]) for c in coldump)

    def set_notify(self, callback=None):
        """Set or clear the notify callback."""
        if callback is not None:
            self.__notify = callback
        else:
            self.__notify = self.__def_notify

    def get_value(self, key):
        """Alternate value fetch."""
        return self.__getitem__(key)

    def set_value(self, key, value):
        """Update a value without triggering notify."""
        key = colkey(key)
        self.__store[key] = value
        if key in ['no', 'series', 'first', 'last', 'org']:
            self.__strcache = {}

    def notify(self):
        """Forced notify."""
        self.__notify(self.get_id())

    def __init__(self, cols={}, no='', series='', notify=None):
        self.__strcache = {}
        self.__store = dict(cols)
        self.__notify = self.__def_notify
        if 'no' not in self.__store:
            self.__store['no'] = no
        if 'series' not in self.__store:
            self.__store['series'] = series
        if notify is not None:
            self.__notify = notify

    def summary(self):
        """Return a summary string for the rider."""
        ret = None
        iv = [
            ' '.join((
                self.get_label(),
                repr(self.get_bibstr()),
                self.listname(16),
            ))
        ]

        colset = _RIDER_COLUMNS
        series = self.__store['series']
        if series in ('cat', 'series'):
            colset = _CATEGORY_COLUMNS
        elif series.startswith('t'):
            colset = _TEAM_COLUMNS

        for k in colset:
            if k not in ('no', 'series', 'first', 'last', 'org'):
                if self[k]:
                    iv.append('%s: %s' % (colset[k], self[k]))

        ret = ', '.join(iv)
        return ret

    def name_bib(self):
        """Return rider name with bib and without org."""
        ret = None
        nkey = 'nb'
        if nkey not in self.__strcache:
            ret = self.__store['no'] + ' ' + self.fitname(48)
            self.__strcache[nkey] = ret
        else:
            ret = self.__strcache[nkey]
        return ret

    def resname_bib(self):
        """Return rider name formatted for results with bib."""
        ret = None
        nkey = 'rnb'
        if nkey not in self.__strcache:
            ret = self.__store['no'] + ' ' + self.listname(48)
            self.__strcache[nkey] = ret
        else:
            ret = self.__strcache[nkey]
        return ret

    def resname(self):
        """Return the name for results"""
        return self.listname(48)

    def listname(self, namelen=32):
        """Return a standard rider name summary field for non-edit lists."""
        ret = None
        nkey = ('ln', namelen)
        if nkey not in self.__strcache:
            ret = self.fitname(namelen)
            if self['org']:
                org = self['org']
                if grapheme.length(org) < 4:
                    org = org.upper()
                ret += ' (' + org + ')'
            self.__strcache[nkey] = ret
        else:
            ret = self.__strcache[nkey]
        return ret

    def fitname(self, width, trunc=False):
        """Return a truncated name string of width or less graphemes"""
        ret = None
        nkey = ('fn', width, trunc)
        if nkey not in self.__strcache:
            if self['series'].startswith('t'):
                ret = self['first']
                if trunc and grapheme.length(ret) > width:
                    if width > 4:
                        ret = grapheme.slice(ret, end=width - 1) + '\u2026'
                    else:
                        ret = grapheme.slice(ret, end=width)
            else:
                fn = self['first'].strip().title()
                fl = grapheme.length(fn)
                ln = self['last'].strip().upper()
                ll = grapheme.length(ln)
                flen = fl + ll
                if fl and ll:
                    flen += 1
                if flen > width:
                    lshrt = ln.split('-')[-1].strip()
                    lsl = grapheme.length(lshrt)
                    flen = fl + lsl
                    if fl and lsl:
                        flen += 1
                    if flen > width and ln:
                        if fl > 2:
                            fshrt = grapheme.slice(fn, end=1) + '.'
                            fn = fshrt
                            fsl = 2
                            flen = fsl + ll
                            if fsl and ll:
                                flen += 1
                    else:
                        ln = lshrt
                if fn and ln:
                    ret = ' '.join((fn, ln))
                elif fn:
                    ret = fn
                else:
                    ret = ln
                if trunc and grapheme.length(ret) > width:
                    if width > 4:
                        ret = grapheme.slice(ret, end=width - 1) + '\u2026'
                    else:
                        ret = grapheme.slice(ret, end=width)
            self.__strcache[nkey] = ret
        else:
            ret = self.__strcache[nkey]
        return ret

    def __str__(self):
        return '{} {} {} ({})'.format(
            strops.bibser2bibstr(self.__store['no'], self.__store['series']),
            self['first'].title(), self['last'].upper(), self['org'].upper())

    def __repr__(self):
        return 'rider({})'.format(self.__store)

    def __len__(self):
        return len(self.__store)

    def __getitem__(self, key):
        """Use a default value id, but don't save it."""
        key = colkey(key)
        if key in self.__store:
            return self.__store[key]
        elif key in _RIDER_DEFAULTS:
            return _RIDER_DEFAULTS[key]
        else:
            return ''

    def __setitem__(self, key, value):
        key = colkey(key)
        self.__store[key] = value
        if key in ['no', 'series', 'first', 'last', 'org']:
            self.__strcache = {}
        self.__notify(self.get_id())

    def __delitem__(self, key):
        key = colkey(key)
        del (self.__store[key])
        if key in ['no', 'series', 'first', 'last', 'org']:
            self.__strcache = {}
        self.__notify(self.get_id())

    def __iter__(self):
        return iter(self.__store.keys())

    def iterkeys(self):
        return iter(self.__store.keys())

    def __contains__(self, item):
        key = colkey(item)
        return key in self.__store

    def __def_notify(self, data=None):
        pass


class riderdb():
    """Rider database."""

    def clear(self, notify=True):
        """Clear rider model."""
        self.__store = {}
        _log.debug('Rider model cleared')
        if notify:
            self.__notify(None)

    def del_rider(self, key, notify=True):
        """Remove rider entry from model with optional notify."""
        del (self.__store[key])
        if notify:
            self.__notify(None)

    def add_rider(self, newrider, notify=True, overwrite=False):
        """Append newrider to model with optional notify."""
        rid = newrider.get_id()
        if rid in self.__store:
            if overwrite:
                olr = self.__store[rid]
                _log.info('Overwriting %s %s', olr.get_label(),
                          olr.resname_bib())
            else:
                _log.info('Duplicate %s %s', newrider.get_label(),
                          newrider.resname_bib())
                rid = (newrider['no'].upper(), '-'.join(
                    ('dupe', strops.randstr())))
        newrider.set_notify(self.__notify)
        self.__store[rid] = newrider
        if notify:
            self.__notify(rid)
        return rid

    def __loadrow(self, r, colspec):
        nr = rider()
        for i in range(0, len(colspec)):
            if len(r) > i:  # column data in row
                key = colspec[i]
                val = cellnorm(r[i])
                if key == 'series':
                    val = val.lower()
                nr[key] = val
        if nr['no']:
            if colkey(nr['no']) in _RIDER_COLUMNS:
                _log.debug('Column header: %r', r)
                return None
        else:
            if nr['series'] != 'series':
                _log.warning('%s without number: %s %s', nr.get_label(),
                             nr.get_bibstr(), nr.resname())
        return nr

    def load(self, csvfile=None, overwrite=False):
        """Load riders from supplied CSV file."""
        if not os.path.isfile(csvfile):
            _log.debug('Riderdb file %r not found', csvfile)
            return 0
        _log.debug('Loading from %r', csvfile)
        count = 0
        with open(csvfile, 'r', encoding='utf-8', errors='replace') as f:
            cr = csv.reader(f)
            incols = None  # no header
            for r in cr:
                nr = None
                if len(r) > 0:  # got a data row
                    if incols is not None:  # already got col header
                        nr = self.__loadrow(r, incols)
                    else:
                        # determine input column structure
                        if colkey(r[0]) in _RIDER_COLUMNS:
                            incols = []
                            for col in r:
                                incols.append(colkey(col))
                        else:
                            incols = _DEFAULT_COLUMN_ORDER  # assume full
                            nr = self.__loadrow(r, incols)
                if nr is not None:
                    self.add_rider(nr, notify=False, overwrite=overwrite)
                    count += 1
        if count > 0:
            self.__notify(None)
        return count

    def listcats(self, series=None):
        """Return a set of categories assigned to riders in the riderdb"""
        if series is not None:
            series = series.lower()
        cats = set()
        for r in self.__store.values():
            if r.in_series(series):
                cats.update(r.get_cats())
        return cats

    def listseries(self):
        """Return an ordered list of series ids in the riderdb"""
        seen = set()
        defined = []
        anonymous = []
        for r in self.__store.values():
            rser = r['series']
            if rser == 'series':
                defined.append(r['no'])
                seen.add(r['no'])
            elif rser not in (
                    'cat',
                    'spare',
                    'ds',
                    'team',
            ):
                if rser not in seen:
                    anonymous.append(rser)
                    seen.add(rser)
        for s in anonymous:
            if s not in defined:
                defined.append(s)
        _log.debug('Returning rider series list: %r', defined)
        return defined

    def biblistfromcat(self, cat, series=None):
        """Return a list of rider ids in the supplied category"""
        if series is not None:
            series = series.lower()
        ret = set()
        label = ''
        for r in self.__store.values():
            if (series is not None
                    and r['series'] == series) or (r['series']
                                                   not in _RESERVED_SERIES):
                if r.in_cat(cat):
                    if not label:
                        label = r.get_label()
                    ret.add(r.get_bibstr())
        _log.debug('Found %d %s in cat %r, series %r', len(ret), label, cat,
                   series)
        return ret

    def biblistfromseries(self, series):
        """Return a list of rider ids in the supplied series"""
        if series is not None:
            series = series.lower()
        ret = set()
        label = ''
        for r in self.__store.values():
            if r['series'] == series:
                if not label:
                    label = r.get_label()
                ret.add(r.get_bibstr())
        _log.debug('Found %d %s in series %r', len(ret), label, series)
        return ret

    def save(self, csvfile=None, columns=None):
        """Save current model content to CSV file."""
        _log.debug('Saving to %r', csvfile)
        cats = []
        if columns is None:
            columns = self.include_cols
        with metarace.savefile(csvfile) as f:
            cr = csv.writer(f, quoting=csv.QUOTE_ALL)
            cr.writerow(get_header(columns))
            for r in self.__store.values():
                if r['series'] in (
                        'cat',
                        'series',
                ):
                    cats.append(r)
                else:
                    cr.writerow(r.get_row(columns))

            if cats:
                cr.writerow(get_header(columns, _CATEGORY_COLUMNS))
                for r in cats:
                    cr.writerow(r.get_row(columns))

    def load_chipfile(self, csvfile=None):
        """Load refids into model from CSV file"""
        _log.debug('Loading refids from %r', csvfile)
        count = 0
        with open(csvfile, 'r', encoding='utf-8', errors='replace') as f:
            cr = csv.reader(f)
            incols = None  # no header
            for r in cr:
                nr = None
                if len(r) > 0:  # got a data row
                    if incols is not None:  # already got col header
                        nr = self.__loadrow(r, incols)
                    else:
                        # determine input column structure
                        if colkey(r[0]) in _RIDER_COLUMNS:
                            incols = []
                            for col in r:
                                incols.append(colkey(col))
                        else:
                            incols = _DEFAULT_COLUMN_ORDER  # assume full
                            nr = self.__loadrow(r, incols)
                if nr is not None:
                    if nr['refid'] and nr['series'] not in _RESERVED_SERIES:
                        lr = self.get_rider(nr['no'], nr['series'])
                        if lr is not None:
                            if nr['refid'] != lr['refid']:
                                lr['refid'] = nr['refid']
                                count += 1
        if count > 0:
            self.__notify(None)
        return count

    def save_chipfile(self, csvfile=None):
        """Save all known refids from model to CSV file"""
        _log.debug('Export chipfile to %r', csvfile)
        columns = ('refid', 'no', 'series', 'first', 'last', 'cat')
        count = 0
        with metarace.savefile(csvfile) as f:
            cr = csv.writer(f, quoting=csv.QUOTE_ALL)
            cr.writerow(get_header(columns))
            for r in self.__store.values():
                if r['series'] not in _RESERVED_SERIES:
                    if r['refid']:
                        cr.writerow(r.get_row(columns))
                        count += 1
        return count

    def update_cats(self, oldcat, newcat, notify=True):
        """Update all instances of oldcat to newcat in each of the riders"""
        for r in self.__store.values():
            if r['series'] != 'cat':
                oldcat = oldcat.upper()
                newcat = newcat.upper()
                rcv = r['cat'].upper().split()
                if oldcat in rcv:
                    oft = rcv.index(oldcat)
                    rcv.insert(oft, newcat)
                    del (rcv[oft + 1])
                    r.set_value('cat', ' '.join(rcv))
                    if notify:
                        r.notify()

    def items(self):
        return self.__store.items()

    def __len__(self):
        return len(self.__store)

    def __getitem__(self, key):
        return self.__store[key]

    def __setitem__(self, key, value):
        self.__store[key] = value  # no change to key
        self.__notify(key)

    def __delitem__(self, key):
        del (self.__store[key])
        self.__notify(None)

    def __iter__(self):
        return self.__store.__iter__()

    def get_cat(self, catid):
        """Return a handle to a category entry, or None"""
        return self.get_rider(catid, 'cat')

    def next_rider(self, riderno, series=None):
        """Try to return next rider from series"""
        if series is not None:
            series = series.lower()
        ret = None
        curid = self.get_id(riderno, series)
        if curid is not None:
            curSeries = curid[1]
            aux = []
            count = 0
            for r in self.__store:
                if curSeries == r[1]:
                    aux.append((strops.bibstr_key(r[0]), count, r))
                    count += 1
            aux.sort()
            i = 0
            rv = None
            while i < len(aux) - 1:
                if aux[i][2] == curid:
                    rv = i + 1
                    break
                i += 1
            if rv is not None:
                ret = aux[rv][2]
        return ret

    def add_empty(self, riderno, series=None):
        """Add a new entry for this rider and series"""
        if series is not None:
            series = series.lower()
        ret = None
        if series is None:
            riderno, series = strops.bibstr2bibser(riderno)
        nr = rider(no=riderno, series=series)
        return self.add_rider(nr)

    def get_rider(self, riderno, series=None):
        """If rider exists, return handle else return None"""
        if series is not None:
            series = series.lower()
        ret = None
        rkey = self.get_id(riderno, series)
        if rkey in self.__store:
            ret = self.__store[rkey]
        return ret

    def get_id(self, riderno, series=None):
        """If rider exists, return id else return None"""
        ret = None
        if series is None:
            riderno, series = strops.bibstr2bibser(riderno)
        if series is not None:
            series = series.lower()
        rkey = (riderno.upper(), series.lower())
        if rkey in self.__store:
            ret = rkey
        return ret

    def iterkeys(self):
        return self.__store.iterkeys()

    def __contains__(self, item):
        return item in self.__store

    def notify(self, data=None):
        """Trigger a manual notify call"""
        self.__notify(data)

    def set_notify(self, cb=None):
        """Set the data change notification callback."""
        if cb is None:
            cb = self.__defnotify
        self.__notify = cb
        for ev in self:
            ev.set_notify(cb)

    def __def_notify(self, data=None):
        """Handle changes in db."""
        _log.debug('Notify: %r', data)

    def __init__(self, racetypes=None):
        """Constructor for the event db."""
        self.__store = {}
        self.__notify = self.__def_notify
        self.include_cols = _DEFAULT_COLUMN_ORDER



# SPDX-License-Identifier: MIT
"""JSON Configuration module.

Sectioned configuration with JSON export and import,
implemented as a dictionary of sections. Each section
contains a dictionary of key/value pairs,
where the key is a unicode string and the value may be any
base type supported by python & JSON, as well as the
metarace.tod types: tod and agg.

Config sections have support for schema-based descriptions
with the following keys:

    attr : (str) config attribute in object
    type : (str) value type, one of:
		'none', 'str', 'tod', 'int', 'float', 'chan', 'bool'
    prompt : (str) Text prompt for option
    subtext : (str) Supplementary text for option
    hint : (str) Tooltip, additional info for option
    places : (int) Decimal places for decimal, float and tod types
    defer : (bool) Defer writing changes to object
    readonly : (bool) Control should not allow editing value
    options : (dict) Map of option keys to displayed values
    default : (misc) Default value for option
    control: (str) edit control type, one of:
		'none' : no control should be displayed
		'section' : no value is associated with the option
		'text' : single line text entry
		'short' : short text entry with supplemental text
		'check' : yes/no checkbox
		'choice' : select box choice of options

"""

import json
import os
import csv
import logging
from decimal import Decimal
from metarace.tod import tod, fromobj, mktod
from metarace.strops import confopt_chan, CHAN_UNKNOWN

_log = logging.getLogger('jsonconfig')
_log.setLevel(logging.DEBUG)


def _config_object(obj):
    """De-serialise tod and Decimal objects from config."""
    if '__tod__' in obj or '__agg__' in obj:
        return fromobj(obj)
    elif '__dec__' in obj:
        return Decimal(obj['value'])
    return obj


class _configEncoder(json.JSONEncoder):
    """Serialise tod and Decimal objects to config."""

    def default(self, obj):
        if isinstance(obj, tod):
            return obj.serialize()
        elif isinstance(obj, Decimal):
            return {'__dec__': 1, 'value': str(obj)}
        return json.JSONEncoder.default(self, obj)


class config:

    def __init__(self, default={}):
        """Create config object with deep copy of the provided default."""
        self.__store = {}
        self.__schema = {}
        for section in default:
            self.__store[section] = {}
            for key in default[section]:
                self.__store[section][key] = default[section][key]

    def __str__(self):
        return json.dumps(self.__store, cls=_configEncoder)

    def __repr__(self):
        return 'config({})'.format(repr(self.__store))

    def add_section(self, section, schema=None):
        """Add section to config, optionally overwriting schema.

        Args:
            section (str): Section identifier.
            schema (dict, optional): Config schema dict.

        Returns:
            dict: Handle to config section dictionary.

        Raises:
            TypeError: If section key is not string.

        """
        if isinstance(section, str):
            if section not in self.__store:
                self.__store[section] = dict()
            if schema is not None:
                self.__schema[section] = schema
            return self.__store[section]
        else:
            raise TypeError('Invalid section key: ' + repr(section))

    def has_section(self, section):
        """Return True if section in config."""
        return section in self.__store

    def has_option(self, section, key):
        """Return True if section and key in config."""
        return section in self.__store and key in self.__store[section]

    def has_value(self, section, key):
        """Return True if section:key in config and value is not None."""
        return self.has_option(section,
                               key) and self.__store[section][key] is not None

    def sections(self):
        for sec in self.__store:
            yield sec

    def options(self, section):
        for opt in self.__store[section]:
            yield opt

    def get(self, section, key):
        return self.__store[section][key]

    def set(self, section, key, value):
        if isinstance(key, str):
            self.__store[section][key] = value
        else:
            raise TypeError('Invalid option key: ' + repr(key))

    def get_decimal(self, section, key, default=None):
        ret = default
        try:
            ret = Decimal(self.get(section, key))
        except Exception:
            pass
        return ret

    def get_float(self, section, key, default=None):
        ret = default
        try:
            ret = float(self.get(section, key))
        except Exception:
            pass
        return ret

    def get_chan(self, section, key, default=None):
        ret = default
        try:
            rv = self.get(section, key)
            if rv is None or rv == '':
                ret = None
            else:
                nv = confopt_chan(rv)
                if nv != CHAN_UNKNOWN:
                    ret = nv
        except Exception:
            pass
        return ret

    def get_posint(self, section, key, default=None):
        ret = default
        try:
            ret = int(self.get(section, key))
            if ret < 0:
                ret = default
        except Exception:
            pass
        return ret

    def get_int(self, section, key, default=None):
        ret = default
        try:
            ret = int(self.get(section, key))
        except Exception:
            pass
        return ret

    def get_str(self, section, key, default=None):
        ret = self.get(section, key)
        if not isinstance(ret, str):
            ret = default
        return ret

    def get_bool(self, section, key, default=False):
        ret = default
        temp = self.get(section, key)
        if isinstance(temp, str):
            if default:
                if temp.lower() in ['no', 'false', '0']:
                    ret = False
            else:
                if temp.lower() in ['yes', 'true', '1']:
                    ret = True
        else:
            ret = bool(temp)
        return ret

    def get_tod(self, section, key, default=None):
        ret = default
        nv = self.get(section, key)
        if isinstance(nv, tod):
            ret = nv
        else:
            nv = mktod(nv)
            if nv is not None:
                ret = nv
        return ret

    def get_value(self, section, key):
        """Return config value according to schema."""
        ret = None
        schema = {}
        if section in self.__schema:
            schema = self.__schema[section]
        if key in schema:
            option = schema[key]
            if 'default' in option:
                ret = option['default']
            # does the schema require a value?
            otype = 'str'
            if 'type' in option:
                otype = option['type']
            ctrl = 'text'
            if 'control' in option:
                ctrl = option['control']
            if otype != 'none' and ctrl != 'section':
                # schema expects a value
                if self.has_option(section, key):
                    if otype == 'tod':
                        ret = self.get_tod(section, key, ret)
                    elif otype == 'int':
                        ret = self.get_int(section, key, ret)
                    elif otype == 'chan':
                        ret = self.get_chan(section, key, ret)
                    elif otype == 'bool':
                        ret = self.get_bool(section, key, ret)
                    elif otype == 'float':
                        ret = self.get_float(section, key, ret)
                    else:
                        # assume str
                        ret = self.get_str(section, key, ret)
                else:
                    pass
                    #_log.debug('Default used for %r:%r', section, key)
            else:
                pass
                #_log.debug('No value assigned for option %r:%r', section,key)
        else:
            _log.debug('Requested value %r:%r not in schema', section, key)
            if self.has_option(section, key):
                ret = self.get(section, key)
        return ret

    def export_section(self, section, obj):
        """Copy values from section to obj according to schema."""
        if section not in self.__schema:
            _log.error('No schema for section export %r', section)
            return False
        for option in self.__schema[section]:
            schema = self.__schema[section][option]
            otype = 'txt'
            if 'type' in schema:
                otype = schema['type']
            if otype != 'none':
                if 'attr' in schema and hasattr(obj, schema['attr']):
                    attr = schema['attr']
                    val = self.get_value(section, option)
                    setattr(obj, attr, val)
                    #_log.debug('Export option:%r, attr:%r, val:%r', option,
                    #           attr, val)
                else:
                    pass
                    #_log.debug('Skip option: %r', option)
            else:
                pass
                #_log.debug('Option: %r type none', option)

    def export_dict(self, section):
        """Return section as dict according to schema."""
        ret = {}
        if section not in self.__schema:
            _log.error('No schema for section export %r', section)
            return None
        for option in self.__schema[section]:
            schema = self.__schema[section][option]
            otype = 'txt'
            if 'type' in schema:
                otype = schema['type']
            if otype != 'none':
                val = self.get_value(section, option)
                ret[option] = val
            else:
                pass
        return ret

    def import_csv(self, filename, defaults=None):
        """Import values from csv into self according to schema."""
        with open(filename, encoding='utf-8') as f:
            section = None
            cr = csv.reader(f)
            for r in cr:
                if len(r) > 1 and r[0]:
                    key = r[0].strip()
                    val = r[1].strip()
                    if key.lower() == 'section':
                        section = val
                        self.add_section(section)
                        _log.debug('CSV import set section to %r', section)
                        if defaults is not None:
                            self.merge(defaults, section)
                    elif section is not None:
                        self.set(section, key, val)
                        nv = self.get_value(section, key)
                        _log.debug('CSV Import key %r: %r => (%s) %r', key,
                                   val, nv.__class__.__name__, nv)
                        # Write back schema-adjusted value
                        self.set(section, key, nv)
                    else:
                        _log.debug('CSV import key %r ignored, no section',
                                   key)

    def import_section(self, section, obj):
        """Copy values from obj into section according to schema."""
        if section not in self.__schema:
            _log.error('No schema for section import %r', section)
            return False
        for option in self.__schema[section]:
            schema = self.__schema[section][option]
            otype = 'txt'
            if 'type' in schema:
                otype = schema['type']
            if otype != 'none':
                if 'attr' in schema and hasattr(obj, schema['attr']):
                    attr = schema['attr']
                    val = getattr(obj, attr)
                    self.set(section, option, val)
                    #_log.debug('Import option:%r, attr:%r, val:%r', option,
                    #attr, val)
                else:
                    pass
                    #_log.debug('Skip option: %r', option)
            else:
                pass
                #_log.debug('Option: %r type none', option)

    def write(self, file):
        """Write config to file."""
        json.dump(self.__store, file, indent=1, cls=_configEncoder)

    def dumps(self):
        """Dump config as JSON string."""
        return json.dumps(self.__store, indent=1, cls=_configEncoder)

    def dictcopy(self):
        """Return a copy of the configuration as a dictionary object."""
        return dict(self.__store)

    def merge(self, otherconfig, section=None, key=None):
        """Merge values from otherconfig into self."""
        if not isinstance(otherconfig, config):
            raise TypeError('Merge expects jsonconfig object.')
        if key is not None and section is not None:
            if otherconfig.has_option(section, key):
                self.set(section, key, otherconfig.get(section, key))
        elif section is not None:
            self.add_section(section)
            if otherconfig.has_section(section):
                for opt in otherconfig.options(section):
                    self.set(section, opt, otherconfig.get(section, opt))
        else:
            for sec in otherconfig.sections():
                if self.has_section(sec):
                    # in this case, only add sections already defined
                    for opt in otherconfig.options(sec):
                        self.set(sec, opt, otherconfig.get(sec, opt))

    def reads(self, s):
        """Read config from a JSON-encoded string."""
        self.addconf(json.loads(s, object_hook=_config_object))

    def read(self, file):
        """Read config from file."""
        self.addconf(json.load(file, object_hook=_config_object))

    def addconf(self, obj):
        """Add all sections and values from obj to self."""
        if not isinstance(obj, dict):
            raise TypeError('Configuration file is not dict: ' +
                            obj.__class__.__name__)
        for sec in obj:
            thesec = obj[sec]
            if not isinstance(thesec, dict):
                raise TypeError('Configuration section is not dict: ' +
                                thesec.__type__.__name__)
            self.add_section(sec)
            for k in thesec:
                self.set(sec, k, thesec[k])

    def load(self, filename):
        """Load the configuration from filename, return True/False."""
        ret = True
        if os.path.exists(filename):
            try:
                with open(filename, mode='rb') as f:
                    self.read(f)
                _log.debug('Loaded from %r', filename)
            except Exception as e:
                _log.error('%s loading config: %s', e.__class__.__name__, e)
                ret = False
        else:
            _log.debug('Load file %r not found', filename)
            ret = False
        return ret



# SPDX-License-Identifier: MIT
"""Tag Heuer/Chronelec Decoder Interface."""

# For connections to multiple decoders, use thbchub

import queue
import logging
import serial
import socket
import time

from . import (decoder, DECODER_LOG_LEVEL)
from metarace import sysconf
from metarace import tod

_log = logging.getLogger('decoder.thbc')
_log.setLevel(logging.DEBUG)

THBC_BAUD = 19200
THBC_UDP_PORT = 2008
THBC_ENCODING = 'iso8859-1'

# THbC protocol messages
ESCAPE = b'\x1b'
HELOCMD = b'MR1'
STOPCMD = ESCAPE + b'\x13\x5c'
REPEATCMD = ESCAPE + b'\x12'
ACKCMD = ESCAPE + b'\x11'

STATCMD = ESCAPE + b'\x05'  # fetch status
CHKCMD = ESCAPE + b'\x06'  # UNKNOWN
STARTCMD = ESCAPE + b'\x07'  # start decoder
SETCMD = ESCAPE + b'\x08'  # set configuration
IPCMD = ESCAPE + b'\x09'  # set IP configuration
QUECMD = ESCAPE + b'\x10'  # fetch configuration

STALVL = ESCAPE + b'\x1e'
BOXLVL = ESCAPE + b'\x1f'

NACK = b'\x07'
CR = b'\x0d'
LF = b'\x0a'
SETTIME = ESCAPE + b'\x48'
STATSTART = b'['
PASSSTART = b'<'

# decoder config consts
IPCONFIG_LEN = 16
CONFIG_LEN = 27
CONFIG_TOD = 0
CONFIG_GPS = 1
CONFIG_TZ_HOUR = 2
CONFIG_TZ_MIN = 3
CONFIG_485 = 4
CONFIG_FIBRE = 5
CONFIG_PRINT = 6
CONFIG_MAX = 7
CONFIG_PROT = 8
CONFIG_PULSE = 9
CONFIG_PULSEINT = 10
CONFIG_CELLSYNC = 11
CONFIG_CELLTOD_HOUR = 12
CONFIG_CELLTOD_MIN = 13
CONFIG_TONE_STA = 15
CONFIG_TONE_BOX = 17
CONFIG_TONE_MAN = 19
CONFIG_TONE_CEL = 21
CONFIG_TONE_BXX = 23
CONFIG_ACTIVE_LOOP = 14
CONFIG_SPARE = 25
CONFIG_FLAGS = {
    CONFIG_TOD: 'Time of Day',
    CONFIG_GPS: 'GPS Sync',
    CONFIG_TZ_HOUR: 'Timezone Hour',
    CONFIG_TZ_MIN: 'Timezone Min',
    CONFIG_485: 'Distant rs485',
    CONFIG_FIBRE: 'Distant Fibre',
    CONFIG_PRINT: 'Serial Print',
    CONFIG_MAX: 'Detect Max',
    CONFIG_PROT: 'Protocol',
    CONFIG_PULSE: 'Sync Pulse',
    CONFIG_PULSEINT: 'Sync Interval',
    CONFIG_CELLSYNC: 'CELL Sync',
    CONFIG_CELLTOD_HOUR: 'CELL Sync Hour',
    CONFIG_CELLTOD_MIN: 'CELL Sync Min',
    CONFIG_TONE_STA: 'STA Tone',
    CONFIG_TONE_BOX: 'BOX Tone',
    CONFIG_TONE_MAN: 'MAN Tone',
    CONFIG_TONE_CEL: 'CEL Tone',
    CONFIG_TONE_BXX: 'BXX Tone',
    CONFIG_ACTIVE_LOOP: 'Active Loop',
    CONFIG_SPARE: '[spare]'
}
DEFPORT = '/dev/ttyS0'

_CONFIG_SCHEMA = {
    'loopsec': {
        'prompt': 'Chronelec/Tag Heuer Protime RC Decoder',
        'control': 'section'
    },
    'activeloop': {
        'prompt': 'Loop Type:',
        'subtext': 'Active?',
        'hint': 'If set, provide power to connected loops',
        'control': 'check',
        'type': 'bool',
        'default': True,
        'index': CONFIG_ACTIVE_LOOP,
    },
    'lvl': {
        'prompt': 'Level:',
        'subtext': '(0 to 99)',
        'hint': 'Accept passing when read level exceeds this value',
        'control': 'short',
        'type': 'int',
    },
    'detectmax': {
        'prompt': 'Detect Max:',
        'subtext': 'Yes?',
        'hint':
        'If set, decoder attempts to report time associated with maximum reading',
        'control': 'check',
        'type': 'bool',
        'default': False,
        'index': CONFIG_MAX,
    },
    'syncsec': {
        'prompt': 'Synchronisation',
        'control': 'section',
    },
    'gpssync': {
        'prompt': 'GPS:',
        'subtext': 'Set time by GPS?',
        'hint': 'If supported by decoder, set time using GPS',
        'control': 'check',
        'type': 'bool',
        'default': False,
        'index': CONFIG_GPS,
    },
    # cell sync only applies to a stopped decoder
    'cellsync': {
        'prompt': 'Sync In:',
        'subtext': 'Start decoder by CELL?',
        'hint':
        'If set, an impulse on CELL will start a stopped decoder with time set to Sync Hour and Sync Min',
        'control': 'none',
        'type': 'bool',
        'default': False,
        'index': CONFIG_CELLSYNC,
    },
    'synchour': {
        'prompt': 'Sync Hour:',
        'subtext': '(0 to 24)',
        'control': 'none',
        'type': 'int',
        'default': 0,
        'index': CONFIG_CELLTOD_HOUR,
    },
    'syncmin': {
        'prompt': 'Sync Min:',
        'subtext': '(0 to 59)',
        'control': 'none',
        'type': 'int',
        'default': 0,
        'index': CONFIG_CELLTOD_MIN,
    },
    'syncpulse': {
        'prompt': 'Trig Out:',
        'subtext': 'Yes?',
        'hint': 'Assert trigger out periodically according to Trig Interval',
        'control': 'check',
        'type': 'bool',
        'default': False,
        'index': CONFIG_PULSE,
    },
    'syncinterval': {
        'prompt': 'Trig Interval:',
        'subtext': '(seconds)',
        'control': 'short',
        'type': 'int',
        'default': 60,
        'index': CONFIG_PULSEINT,
    },
    'beepsec': {
        'prompt': 'Audio Tone Configuration',
        'control': 'section',
    },
    'statone': {
        'prompt': 'STA Tone:',
        'subtext': '(Hz)',
        'control': 'short',
        'type': 'int',
        'default': 514,
        'index': CONFIG_TONE_STA,
    },
    'boxtone': {
        'prompt': 'BOX Tone:',
        'subtext': '(Hz)',
        'control': 'short',
        'type': 'int',
        'default': 770,
        'index': CONFIG_TONE_BOX,
    },
    'mantone': {
        'prompt': 'MAN Tone:',
        'subtext': '(Hz)',
        'control': 'short',
        'type': 'int',
        'default': 686,
        'index': CONFIG_TONE_MAN,
    },
    'celltone': {
        'prompt': 'CELL Tone:',
        'subtext': '(Hz)',
        'control': 'short',
        'type': 'int',
        'default': 611,
        'index': CONFIG_TONE_CEL,
    },
    'bxxtone': {
        'prompt': 'BXX Tone:',
        'subtext': '(Hz)',
        'control': 'short',
        'type': 'int',
        'default': 915,
        'index': CONFIG_TONE_BXX,
    },
    'hwsec': {
        'prompt': 'Hardware Options',
        'hint': 'These options are not used',
        'control': 'none',
    },
    'protocol': {
        'prompt': 'Protocol:',
        'subtext': '(undocumented)',
        'control': 'none',
        'type': 'int',
        'default': 0,
        'index': CONFIG_PROT,
    },
    'tod': {
        'prompt': 'Time of Day:',
        'subtext': 'Yes?',
        'hint': 'Required for normal use',
        'control': 'none',
        'type': 'bool',
        'default': True,
        'index': CONFIG_TOD,
    },
    'tzhr': {
        'prompt': 'Time Zone:',
        'subtext': '(hour)',
        'control': 'none',
        'type': 'int',
        'default': 0,
        'index': CONFIG_TZ_HOUR,
    },
    'tzmin': {
        'prompt': 'Time Zone:',
        'subtext': '(min)',
        'control': 'none',
        'type': 'int',
        'default': 0,
        'index': CONFIG_TZ_MIN,
    },
    'serprint': {
        'prompt': 'Serial Print:',
        'subtext': 'Yes?',
        'hint': 'Serial printer connected on aux port',
        'control': 'none',
        'type': 'bool',
        'default': False,
        'index': CONFIG_PRINT,
    },
    'distfibre': {
        'prompt': 'Distant Decoder:',
        'subtext': 'Fibre?',
        'hint': 'Distant decoder connected by fibre optic link',
        'control': 'none',
        'type': 'bool',
        'default': False,
        'index': CONFIG_FIBRE,
    },
    'dist485': {
        'prompt': 'Distant Decoder:',
        'subtext': 'RS-485?',
        'hint': 'Distant decoder connected via RS-485',
        'control': 'none',
        'type': 'bool',
        'default': False,
        'index': CONFIG_485,
    },
    'ipconf': {
        'prompt': 'IP Configuration',
        'control': 'section',
    },
    'ipaddr': {
        'prompt': 'IP:',
        'hint': 'IP address of decoder',
        'default': '',
    },
    'netmask': {
        'prompt': 'Netmask:',
        'hint': 'Netmask',
        'default': '',
    },
    'host': {
        'prompt': 'Host:',
        'hint': 'Address decoder reports to',
        'default': '',
    },
    'gateway': {
        'prompt': 'Gateway:',
        'hint': 'Gateway IP',
        'default': '',
    },
}


def reflect(dat, width):
    """Reverse bit order of byte"""
    out = dat & 0x01
    for i in range(width - 1):
        dat >>= 1
        out = (out << 1) | (dat & 0x01)
    return out


# Build a lookup table for the MCRF4XX CRC
# Source: pycrc https://pycrc.org/
_MCRF4XXTBL = [0] * 256
for i in range(256):
    r = i
    r = reflect(r, 8) << 8
    for j in range(8):
        if r & 0x8000 != 0:
            r = (r << 1) ^ 0x1021
        else:
            r = (r << 1)
    r = reflect(r, 16)
    _MCRF4XXTBL[i] = r & 0xffff


def mcrf4xx(msgstr=b''):
    """Return MCRF4XX CRC for provided byte string."""
    r = 0xffff  # reflect(0xffff,16) == 0xffff
    for b in msgstr:
        i = (r ^ b) & 0xff
        r = ((r >> 8) ^ _MCRF4XXTBL[i]) & 0xffff
    return r & 0xffff  # collapse two reflects, mask and ^0


def thbc_sum(msgstr=b''):
    """Return sum of character values as decimal string."""
    ret = 0
    for ch in msgstr:
        ret = ret + ch
    return '{0:04d}'.format(ret).encode('ascii', 'ignore')


def val2hexval(val):
    """Convert int to decimal digit equivalent hex byte."""
    ret = 0x00
    ret |= ((val // 10) & 0x0f) << 4  # msd	97 -> 0x90
    ret |= (val % 10) & 0x0f  # lsd   97 -> 0x07
    return ret


def hexval2val(hexval):
    """Unconvert a decimal digit equivalent hex byte to int."""
    ret = 10 * (hexval >> 4)  # tens 0x97 -> 90
    ret += hexval & 0x0f  # ones 0x97 ->  7
    return ret


class thbc(decoder):
    """Tag Heuer / Chronelec thread object class."""

    def __init__(self):
        decoder.__init__(self)
        self._boxname = None
        self._version = ''
        self._decoderconfig = {}
        self._decoderipconfig = {}
        self._io = None
        self._cksumerr = 0
        self._rdbuf = b''  # bytestring read buffer
        self._curport = None

    # API overrides
    def status(self):
        """Request status message from decoder."""
        self.write(STATCMD)

    def connected(self):
        return self._io is not None and self._boxname is not None

    def stop_session(self):
        """Send a stop command to decoder."""
        self.write(STOPCMD)

    def start_session(self):
        """Send a depart command to decoder."""
        self.write(STARTCMD)

    def clear(self):
        """Start a new session and request time sync."""
        self.stop_session()
        self.start_session()
        self.sync()

    def get_config(self):
        """Request decoder configuration."""
        self.write(QUECMD)

    def ipconfig(self):
        """Request sanity check in decoder thread."""
        self._cqueue.put_nowait(('_ipcfg', None))

    # Device-specific functions
    def _close(self):
        if self._io is not None:
            _log.debug('Close connection')
            cp = self._io
            self._io = None
            try:
                cp.close()
            except Exception as e:
                _log.debug('%s closing io: %s', e.__class__.__name__, e)

    def _port(self, port):
        """Re-establish connection to supplied device port."""
        self._close()
        if port is None and self._curport is not None:
            port = self._curport
        if port is None:
            _log.debug('Re-connect cancelled: port is None')
            return
        s = None
        self._rdbuf = b''
        if '/' not in port and '.' in port:
            _log.debug('Attempting UDP on %r', port)
            s = dgram(port, THBC_UDP_PORT)
        else:
            # assume device file
            s = serial.Serial(port, THBC_BAUD, rtscts=False, timeout=0.2)
        self._boxname = None
        self._io = s
        self._write(QUECMD)
        # queue sane through command loop
        time.sleep(0.2)
        self.sane()
        self._curport = port

    def _sync(self, data=None):
        _log.debug('Performing blocking sync')
        acceptval = tod.tod('0.001')
        nt = tod.now()
        diff = nt - nt.truncate(0)
        while diff > acceptval and diff < tod.ONE:
            time.sleep(0.0005)
            nt = tod.now()
            diff = nt - nt.truncate(0)
        self._write(self._set_time_cmd(nt))
        _log.debug('Set time: %r', nt.meridiem())

    def _ipcfg(self, data=None):
        """Alter the attached decoder's IP address."""
        sysconf.add_section('thbc', _CONFIG_SCHEMA)
        cmd = b'\x09\x09'
        defval = socket.inet_aton(socket.gethostbyname(''))
        for opt in ['ipaddr', 'netmask', 'gateway', 'host']:
            dv = defval
            try:
                nv = sysconf.get_value('thbc', opt)
                hn = socket.gethostbyname(nv)
                _log.debug('Updating config %s %r',
                           _CONFIG_SCHEMA[opt]['prompt'], hn)
                dv = socket.inet_aton(hn)
            except Exception as e:
                _log.info('%s reading IP config %r: %s', e.__class__.__name__,
                          opt, e)
            cmd += dv
        _log.info('Starting IP config update')
        self._v3_cmd(cmd)

    def _sane(self, data=None):
        """Check decoder config against system settings."""
        sysconf.add_section('thbc', _CONFIG_SCHEMA)
        doconf = False
        if self._boxname is not None:
            for key in _CONFIG_SCHEMA:
                if 'index' in _CONFIG_SCHEMA[key]:
                    nv = sysconf.get_value('thbc', key)
                    if nv is not None:
                        flag = _CONFIG_SCHEMA[key]['index']
                        if self._decoderconfig[flag] != nv:
                            _log.debug('Config key %r differs: %r != %r', key,
                                       self._decoderconfig[flag], nv)
                            self._decoderconfig[flag] = nv
                            doconf = True
        else:
            _log.info('Decoder not connected')

        # re-write config if required
        if doconf:
            _log.info('Updating %r config', self._boxname)
            self._set_config()

        # force decoder threshold level
        lvl = sysconf.get_value('thbc', 'lvl')
        if lvl is not None:
            _log.info('Setting detection levels to: %r', lvl)
            self._setlvl(lvl)

    def _v3_cmd(self, cmdstr=b''):
        """Pack and send a v3 command directly to port."""
        crc = mcrf4xx(cmdstr)
        crcstr = bytes([(crc >> 8) & 0xff, crc & 0xff])
        self._write(ESCAPE + cmdstr + crcstr + b'>')

    def _serialise_config(self):
        """Convert current decoder setting into a config string"""
        obuf = bytearray(CONFIG_LEN)
        # fill in level bytes
        obuf[CONFIG_SPARE] = 0x13  # will be fixed by subsequent levelset
        obuf[CONFIG_SPARE + 1] = 0x15

        # fill in tone values
        for opt in [
                CONFIG_TONE_STA, CONFIG_TONE_BOX, CONFIG_TONE_MAN,
                CONFIG_TONE_CEL, CONFIG_TONE_BXX
        ]:
            if opt in self._decoderconfig:
                obuf[opt] = val2hexval(self._decoderconfig[opt] // 100)  # xx00
                obuf[opt + 1] = val2hexval(self._decoderconfig[opt] %
                                           100)  # 00xx

        # fill in single byte values
        for opt in [
                CONFIG_TZ_HOUR, CONFIG_TZ_MIN, CONFIG_PROT, CONFIG_PULSEINT,
                CONFIG_CELLTOD_HOUR, CONFIG_CELLTOD_MIN
        ]:
            if opt in self._decoderconfig:
                obuf[opt] = val2hexval(self._decoderconfig[opt] % 100)  # ??

        # fill in flags
        for opt in [
                CONFIG_TOD, CONFIG_GPS, CONFIG_485, CONFIG_FIBRE, CONFIG_PRINT,
                CONFIG_MAX, CONFIG_PULSE, CONFIG_CELLSYNC, CONFIG_ACTIVE_LOOP
        ]:
            if opt in self._decoderconfig:
                if self._decoderconfig[opt]:
                    obuf[opt] = 0x01
        return bytes(obuf)

    def _set_config(self):
        """Write current configuration to decoder."""
        cmd = b'\x08\x08' + self._serialise_config()
        self._v3_cmd(cmd)
        self._write(QUECMD)

    def _set_date(self, timestruct=None):
        """Set the date on the decoder."""
        if timestruct is None:
            timestruct = time.localtime()
        _log.debug('Set date on decoder: %s',
                   time.strftime('%Y-%m-%d', timestruct))
        cmd = bytearray(5)
        cmd[0] = 0x0a
        cmd[1] = 0x0a
        cmd[2] = 0xff & timestruct[2]  # day
        cmd[3] = 0xff & timestruct[1]  # month
        cmd[4] = 0xff & (timestruct[0] - 2000)  # year, after 2000
        self._v3_cmd(bytes(cmd))

    def _setlvl(self, lvl=10):
        """Set the read level on box and sta channels."""
        lstr = '%02d' % (max(0, min(99, lvl)))
        self.write(BOXLVL + lstr.encode(THBC_ENCODING))
        self.write(STALVL + lstr.encode(THBC_ENCODING))

    def _set_time_cmd(self, t):
        """Return a set time command string for the provided time of day."""
        body = bytearray(4)
        s = int(t.timeval)
        body[0] = s // 3600  # hours
        body[1] = (s // 60) % 60  # minutes
        body[2] = s % 60  # seconds
        body[3] = 0x74
        return SETTIME + bytes(body)

    def _parse_config(self, msg):
        # decoder configuration message.
        ibuf = bytearray(msg)
        self._decoderconfig = {}
        for flag in sorted(CONFIG_FLAGS):  # import all
            # tone values
            if flag in [
                    CONFIG_TONE_STA, CONFIG_TONE_BOX, CONFIG_TONE_MAN,
                    CONFIG_TONE_CEL, CONFIG_TONE_BXX
            ]:
                self._decoderconfig[flag] = 100 * hexval2val(ibuf[flag])
                self._decoderconfig[flag] += hexval2val(ibuf[flag + 1])

            # single byte values
            elif flag in [
                    CONFIG_TZ_HOUR, CONFIG_TZ_MIN, CONFIG_PROT,
                    CONFIG_PULSEINT, CONFIG_CELLTOD_HOUR, CONFIG_CELLTOD_MIN
            ]:
                self._decoderconfig[flag] = hexval2val(ibuf[flag])

            # 'booleans'
            elif flag in [
                    CONFIG_TOD, CONFIG_GPS, CONFIG_485, CONFIG_FIBRE,
                    CONFIG_PRINT, CONFIG_MAX, CONFIG_PULSE, CONFIG_CELLSYNC,
                    CONFIG_ACTIVE_LOOP
            ]:
                self._decoderconfig[flag] = bool(ibuf[flag])

        self._boxname = ''
        for c in msg[43:47]:
            self._boxname += chr(c + ord('0'))
        self._version = str(hexval2val(ibuf[47]))
        stalvl = hexval2val(msg[25])
        boxlvl = hexval2val(msg[26])
        _log.info('Info Decoder ID: %s', self._boxname)
        _log.debug('Info Firmware Version: %r', self._version)
        _log.debug('Levels: STA=%r, BOX=%r', stalvl, boxlvl)
        self._decoderipconfig['IP'] = socket.inet_ntoa(msg[27:31])
        self._decoderipconfig['Mask'] = socket.inet_ntoa(msg[31:35])
        self._decoderipconfig['Gateway'] = socket.inet_ntoa(msg[35:39])
        self._decoderipconfig['Host'] = socket.inet_ntoa(msg[39:43])
        for key in ['IP', 'Mask', 'Host', 'Gateway']:
            _log.debug('%r: %r', key, self._decoderipconfig[key])

    def _parse_message(self, msg, ack=True):
        """Return tod object from timing msg or None."""
        ret = None
        if len(msg) > 4:
            if msg[0] == PASSSTART[0]:  # RFID message
                idx = msg.find(b'>')
                if idx == 37:  # Valid length
                    data = msg[1:33]
                    msum = msg[33:37]
                    tsum = thbc_sum(data)
                    if tsum == msum:  # Valid 'sum'
                        pvec = data.decode(THBC_ENCODING).split()
                        istr = pvec[3] + ':' + pvec[5]
                        rstr = pvec[1].lstrip('0')
                        cstr = 'C1'
                        if pvec[0] == 'BOX':
                            cstr = 'C2'
                        elif pvec[0] == 'MAN':
                            cstr = 'C0'
                        if pvec[5] == '2':  # LOW BATTERY ALERT
                            _log.warning('Low battery on %r', rstr)
                        elif pvec[5] == '3':  # BATTERY FAILURE ALERT
                            _log.warning('Faulty battery on %r', rstr)
                        ret = tod.tod(pvec[2],
                                      index=istr,
                                      chan=cstr,
                                      refid=rstr,
                                      source=self._boxname)
                        # Log a hardware-specific passing
                        _log.log(DECODER_LOG_LEVEL, msg.strip())
                        if ack:
                            self._write(ACKCMD)  # Acknowledge if ok
                        self._cksumerr = 0
                    else:
                        _log.warning('Invalid checksum: %r != %r: %r', tsum,
                                     msum, msg)
                        self._cksumerr += 1
                        if self._cksumerr > 3:
                            # assume error on decoder, so acknowledge and
                            # continue with log
                            # NOTE: This path is triggered when serial comms
                            # fail and a tag read happens before a manual trig
                            _log.error('Erroneous message from decoder')
                            if ack:
                                self._write(ACKCMD)
                else:
                    _log.debug('Invalid message: %r', msg)
            elif msg[0] == STATSTART[0]:  # Status message
                data = msg[1:22]
                pvec = data.decode(THBC_ENCODING).split()
                if len(pvec) == 5:
                    _log.info('%r@%s Noise:%s/%s Levels:%s/%s', self._boxname,
                              pvec[0], pvec[1], pvec[2], pvec[3], pvec[4])
                else:
                    _log.info('Invalid status: %r', msg)
            elif b'+++' == msg[0:3] and len(msg) > 53:
                self._parse_config(msg[3:])
            else:
                pass
        else:
            _log.debug('Short message: %r', msg)
        return ret

    def _ipcompletion(self):
        """Blocking wait for ipconfig completion - horrible."""
        _log.info('IP Config')
        time.sleep(10)
        self.write(QUECMD)

    def _read(self):
        """Read messages from the decoder until a timeout condition."""
        ch = self._io.read(1)
        while ch != b'':
            if ch == LF and len(self._rdbuf) > 0 and self._rdbuf[-1] == CR[0]:
                # Return ends the current 'message', if preceeded by return
                self._rdbuf += ch  # include trailing newline
                #_log.debug('RECV: %r', self._rdbuf)
                t = self._parse_message(self._rdbuf.lstrip(b'\0'))
                if t is not None:
                    self._trig(t)
                self._rdbuf = b''
            elif len(self._rdbuf) > 40 and b'\x1e\x86\x98' in self._rdbuf:
                # Assume acknowledge from IP Command
                #_log.debug('RECV: %r', self._rdbuf)
                self._rdbuf = b''
                self._ipcompletion()
            else:
                self._rdbuf += ch
            ch = self._io.read(1)

    def _write(self, msg):
        if self._io is not None:
            self._io.write(msg)
            #_log.debug('SEND: %r', msg)

    def run(self):
        """Decoder main loop."""
        _log.debug('Starting')
        self._running = True
        while self._running:
            try:
                c = None
                if self._io is not None:
                    # Read responses until response complete or timeout
                    try:
                        self._read()
                    except socket.timeout:
                        pass
                    c = self._cqueue.get_nowait()
                else:
                    c = self._cqueue.get()
                self._cqueue.task_done()
                self._proccmd(c)
            except queue.Empty:
                pass
            except (serial.SerialException, socket.error) as e:
                self._close()
                self._boxname = None
                _log.error('%s: %s', e.__class__.__name__, e)
            except Exception as e:
                _log.critical('%s: %s', e.__class__.__name__, e)
                self._boxname = None
                self._running = False
        self.setcb()
        _log.debug('Exiting')


class dgram:
    """Serial-like UDP port object."""

    def __init__(self, host, port):
        self._host = host
        self._port = port
        self._s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._s.settimeout(0.2)
        self._s.bind(('', self._port))
        self._buf = b''

    def read(self, sz=1):
        ret = b''
        if len(self._buf) == 0:
            nb, addr = self._s.recvfrom(4096)  # timeout raises exception
            if addr[0] == self._host:
                self._buf += nb
        if len(self._buf) > 0:
            # make sure ret is bytes
            ret = self._buf[0:1]
            self._buf = self._buf[1:]
        return ret

    def write(self, buf=b''):
        return self._s.sendto(buf, (self._host, self._port))

    def close(self):
        self._s.close()



# SPDX-License-Identifier: MIT
"""export result files in a subprocess"""

import threading
import subprocess
import logging
import os

import metarace

# Default Rsync options:
#  -a archive mode
#  -z compress file data during the transfer
#  -L transform symlink into referent file/dir
_RSYNC_OPTS = '-azL'

# Password file for TCP daemon connections
_RSYNC_PWD = 'rsync.pwd'

_log = logging.getLogger('export')
_log.setLevel(logging.DEBUG)

_CONFIG_SCHEMA = {
    'ttype': {
        'prompt': 'Result Export',
        'control': 'section',
    },
    'method': {
        'prompt': 'Method:',
        'hint': 'Export mechanism',
        'control': 'choice',
        'options': {
            'ssh': 'Rsync over SSH',
            'rsync': 'Rsync TCP daemon',
            'shell': 'Export script',
        },
    },
    'timeout': {
        'prompt': 'Timeout:',
        'subtext': 'seconds',
        'hint': 'Export command timeout',
        'control': 'short',
        'type': 'int',
        'default': 30
    },
    'host': {
        'prompt': 'Hostname:',
        'hint': 'Optional hostname, IP or SSH host to connect to',
    },
    'port': {
        'prompt': 'Port:',
        'subtext': '(TCP daemon only)',
        'hint': 'Optional TCP port for rsync TCP daemon',
        'type': 'int',
        'control': 'short',
    },
    'username': {
        'prompt': 'Username:',
        'hint': 'Optional username on remote host',
    },
    'basepath': {
        'prompt': 'Basepath:',
        'hint': 'Optional base path on remote server (without trailing slash)',
    }
}


class mirror(threading.Thread):
    """Mirror thread object class."""

    def __init__(self,
                 callback=None,
                 callbackdata=None,
                 localpath=None,
                 remotepath=None,
                 mirrorcmd=None):
        """Construct mirror thread object."""
        threading.Thread.__init__(self, daemon=True)
        self.__cb = None
        if callback is not None:
            self.__cb = callback
        self.__cbdata = None
        if callbackdata is not None:
            self.__cbdata = callbackdata
        self.__localpath = '.'
        if localpath is not None:
            self.__localpath = localpath
        self.__remotepath = ''
        if remotepath is not None:
            self.__remotepath = remotepath
        self.__mirrorcmd = ''
        if isinstance(mirrorcmd, str) and mirrorcmd:
            self.__mirrorcmd = mirrorcmd

        # read configuration from sysconf
        metarace.sysconf.add_section('export', _CONFIG_SCHEMA)
        self.__method = metarace.sysconf.get_value('export', 'method')
        self.__timeout = metarace.sysconf.get_value('export', 'timeout')
        self.__host = metarace.sysconf.get_value('export', 'host')
        self.__port = metarace.sysconf.get_value('export', 'port')
        self.__username = metarace.sysconf.get_value('export', 'username')
        self.__basepath = metarace.sysconf.get_value('export', 'basepath')

        # save return code and output
        self.returncode = None
        self.stderr = None

    def set_remotepath(self, pathstr):
        """Set or clear the remote path value."""
        self.__remotepath = pathstr

    def set_localpath(self, pathstr):
        """Set or clear the local path value."""
        self.__localpath = pathstr

    def set_cb(self, func=None, cbdata=None):
        """Set or clear the event callback."""
        if func is not None:
            self.__cb = func
            self.__cbdata = cbdata
        else:
            self.__cb = None
            self.__cbdata = None

    def run(self):
        """Called via threading.Thread.start()."""
        if self.__method is None:
            _log.info('Export method not set, files not mirrored')
            return None

        # prepare command line
        _log.debug('Starting export method=%r', self.__method)
        if self.__method == 'shell':
            if self.__mirrorcmd:
                arglist = (self.__mirrorcmd, self.__remotepath)
            else:
                _log.warning('No mirrorcmd specified, export cancelled')
                return None
        elif self.__method == 'ssh':
            dest = self.__remotepath
            if self.__basepath:
                dest = os.path.join(self.__basepath, dest)
            if self.__host:
                dest = self.__host + ':' + dest
            if self.__username:
                dest = self.__username + '@' + dest
            arglist = ('rsync', _RSYNC_OPTS, self.__localpath, dest)
        elif self.__method == 'rsync':
            if self.__host:
                pwarg = '--password-file=' + metarace.default_file('rsync.pwd')
                host = self.__host
                if self.__port is not None:
                    host = host + ':' + str(self.__port)
                if self.__username is not None:
                    host = self.__username + '@' + host
                host = 'rsync://' + host
                dest = self.__remotepath
                if self.__basepath:
                    dest = os.path.join(self.__basepath, dest)
                dest = host + '/' + dest
                arglist = ('rsync', pwarg, _RSYNC_OPTS, self.__localpath, dest)
            else:
                _log.info('TCP host missing, export ignored')
                return None
        else:
            _log.info('Unknown export method %r ignored', self.__method)
            return None

        ret = None
        try:
            _log.debug('Calling subprocess: %r', arglist)
            res = subprocess.run(arglist,
                                 timeout=self.__timeout,
                                 check=True,
                                 capture_output=True)
            if self.__cb is not None:
                self.__cb(res, self.__cbdata)
            self.returncode = res.returncode
            ret = res.returncode
        except subprocess.CalledProcessError as e:
            _log.debug('%r stderr: %r', self.__method, e.stderr)
            _log.error('Export command failed with error: %r', e.returncode)
            self.stderr = e.stderr
        except subprocess.TimeoutExpired as e:
            _log.debug('%r stderr: %r', self.__method, e.stderr)
            _log.error('Timeout waiting for export command to complete')
        except Exception as e:
            _log.error('%s: %s', e.__class__.__name__, e)
        _log.debug('Complete: Returned %r', ret)



# SPDX-License-Identifier: MIT
"""A collection of tools for preparing cycle race results."""

import os
import logging
import errno
from tempfile import NamedTemporaryFile
from shutil import copyfile
from metarace import jsonconfig
try:
    from importlib.resources import files, as_file
except ImportError:
    print('Python >= 3.9 required to use this module')
flockstyle = None
try:
    import fcntl
    flockstyle = 'o-g'
except ImportError:
    pass
if flockstyle is None:
    # possibly a windows machine
    import msvcrt
    flockstyle = 'deviant'

VERSION = '2.1.11'
DATA_PATH = os.path.realpath(
    os.path.expanduser(os.path.join('~', 'Documents', 'metarace')))
DEFAULTS_PATH = os.path.join(DATA_PATH, 'default')
RESOURCE_PKG = 'metarace.data'
LOGO = 'metarace_icon.svg'
SYSCONF = 'metarace.json'
PDF_TEMPLATE = 'pdf_template.json'
PROGRAM_TEMPLATE = 'program_template.json'
LOGFILEFORMAT = '%(asctime)s %(levelname)s:%(name)s: %(message)s'
LOGFORMAT = '%(levelname)s %(name)s: %(message)s'
LOGLEVEL = logging.DEBUG  # default console log level
sysconf = jsonconfig.config()  # system-defaults, populated by init() method
_log = logging.getLogger('metarace')
_log.setLevel(logging.DEBUG)


def init():
    """Shared metarace program initialisation."""
    copyconf = mk_data_path()

    # set global logging options
    logging._srcfile = None
    logging.logThreads = 0
    logging.logProcesses = 0

    # read in a system configuration
    conffile = default_file(SYSCONF)
    try:
        if os.path.exists(conffile):
            sysconf.load(conffile)
            # don't copy path-specific config into defaults
            copyconf = False
        else:
            _log.info('System defaults not present, using package defaults')
            ref = files(RESOURCE_PKG).joinpath(SYSCONF)
            with ref.open('rb') as f:
                sysconf.read(f)
            copyconf = True
    except Exception as e:
        _log.error('%s reading system config: %s', e.__class__.__name__, e)

    # if required, create new system default file
    if copyconf:
        _log.info('Creating default system config %s', SYSCONF)
        with savefile(os.path.join(DEFAULTS_PATH, SYSCONF), perm=0o600) as f:
            sysconf.write(f)


def mk_data_path():
    """Create shared data path if it does not yet exist."""
    ret = False
    if not os.path.exists(DATA_PATH):
        _log.info('Creating data directory: %r', DATA_PATH)
        os.makedirs(DATA_PATH)
    if not os.path.exists(DEFAULTS_PATH):
        _log.info('Creating system defaults directory: %r', DEFAULTS_PATH)
        os.makedirs(DEFAULTS_PATH)
        ret = True  # flag copy of config back to defaults path
    lfile = os.path.join(DEFAULTS_PATH, LOGO)
    if not os.path.exists(lfile):
        _log.info('Saving default app logo into defaults path')
        ref = files(RESOURCE_PKG).joinpath(LOGO)
        with ref.open('rb') as sf:
            with savefile(lfile, mode='b') as df:
                df.write(sf.read())
    return ret


def config_path(configpath=None):
    """Return a writeable meet configuration path.

    Args:
        configpath (str): Filename or path.

    Returns:
        str: Path to writeable meet folder or None if not available.

    """
    ret = None
    if configpath is not None:
        ret = configpath
        if not os.path.isdir(ret) and os.path.isfile(ret):
            ret = os.path.dirname(ret)  # assume dangling path contains file
        ret = os.path.realpath(ret)
        _log.debug('Checking for meet %r using %r', configpath, ret)
        if not os.path.exists(ret):
            try:
                _log.info('Creating meet folder %r', ret)
                os.makedirs(ret)
            except Exception as e:
                _log.error('Unable to create folder %r: %s', ret, e)
                ret = None
        if ret is not None:
            try:
                _log.debug('Checking folder %r for write access', ret)
                with NamedTemporaryFile(dir=ret, prefix='.chkwrite_') as f:
                    pass
            except Exception as e:
                _log.error('Unable to access meet folder %r: %s', ret, e)
                ret = None
    return ret


def default_file(filename=''):
    """Return a path to the named file.

    Path components are stripped, then the the following locations
    are checked in order to find the first instance of filename:
        - current working directory
        - DEFAULTS_PATH

    If file is not found, the original filename is returned.

    Args:
        filename (str): Filename to look up.

    Returns:
        str: Full path to filename in defaults or current directory.

    """
    basefile = os.path.basename(filename)
    if basefile in ['..', '.', '', None]:
        _log.debug('Invalid filename %r ignored', filename)
        return None
    ret = basefile
    if os.path.exists(basefile):
        pass
    else:
        try:
            checkfile = os.path.join(DEFAULTS_PATH, basefile)
            os.stat(checkfile)
            ret = checkfile
        except Exception as e:
            # ignore file not found and path errors
            pass
    return ret


def resource_text(name='', encoding='utf-8'):
    """Return string content of named resource.

    Args:
        name (str): Resource name.
        encoding (str, optional): Text file encoding

    Returns:
        str: Text content of named resource.

    Raises:
        FileNotFoundError: If name not available.

    """
    basefile = os.path.basename(name)
    if basefile in ['..', '.', '', None]:
        raise FileNotFoundError('Invalid resource name: ' + repr(name))
    t = files(RESOURCE_PKG).joinpath(basefile)
    if t is not None and t.is_file():
        return t.read_text(encoding=encoding)
    else:
        raise FileNotFoundError('Named resource not found: ' + repr(name))


def resource_file(name=''):
    """Return temporary filename context manager for a named resource.

    Note: Returns context manager for (potentially) temporary filename:

        with resource_file('resource.svg') as filename:
            with open(filename) as file:
                ...

    Args:
        name (str): Resource name.

    Raises:
        FileNotFoundError: If name not available.

    """
    basefile = os.path.basename(name)
    if basefile in ['..', '.', '', None]:
        raise FileNotFoundError('Invalid resource name: ' + repr(name))
    t = files(RESOURCE_PKG).joinpath(basefile)
    if t is not None and t.is_file():
        return as_file(t)
    else:
        raise FileNotFoundError('Named resource not found: ' + repr(name))


class savefile:
    """Tempfile-backed save file contextmanager.

    Create a temporary file with the desired mode and
    encoding and return a context manager and writable
    file handle.

    On close the temporary file is moved to the provided
    filename, or copied if os.rename is not possible.
    """

    def __init__(self,
                 filename,
                 mode='t',
                 encoding='utf-8',
                 tempdir='.',
                 perm=0o644):
        self.__sfile = filename
        self.__path = tempdir
        self.__perm = perm
        if mode == 'b':
            encoding = None
        self.__tfile = NamedTemporaryFile(mode='w' + mode,
                                          suffix='.tmp',
                                          prefix='sav_',
                                          dir=self.__path,
                                          encoding=encoding,
                                          delete=False)

    def __enter__(self):
        return self.__tfile

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__tfile.close()
        if exc_type is not None:
            return False  # raise exception
        # otherwise, file is saved ok in temp file
        os.chmod(self.__tfile.name, self.__perm)
        try:
            os.rename(self.__tfile.name, self.__sfile)
        except OSError as e:
            copyfile(self.__tfile.name, self.__sfile)
            os.unlink(self.__tfile.name)
        return True


def lockpath(path):
    """Request advisory lockfile in path."""
    if flockstyle is None:
        _log.error('Path locking not available')
        return None
    lockfile = None
    filename = os.path.join(path, '.lock')
    try:
        lockfile = open(filename, 'a+b')
        if flockstyle == 'o-g':
            fcntl.flock(lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
            _log.debug('Path lock %r acquired by fcntl', filename)
        else:
            lockfile.seek(0)
            msvcrt.locking(lockfile.fileno(), msvcrt.LK_NBLCK, 1)
            _log.debug('Path lock %r acquired by msvcrt', filename)
    except Exception as e:
        if lockfile is not None:
            lockfile.close()
            lockfile = None
        _log.error('Unable to acquire path lock %r: %s', filename, e)
    return lockfile


def unlockpath(path, lockfile):
    """Release advisory lockfile."""
    filename = os.path.join(path, '.lock')
    if os.path.exists(filename):
        os.unlink(filename)
    lockfile.close()
    _log.debug('Path lock %r released', filename)
    return None


LICENSETEXT = """
MIT License

Copyright (c) 2012-2025 ndf-zz and contributors

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""



# SPDX-License-Identifier: MIT
"""Race Result USB Decoder interface."""

import threading
import queue
import decimal
import logging
import serial
from time import sleep

from . import (decoder, DECODER_LOG_LEVEL)
from metarace import tod
from metarace import sysconf
from metarace import strops

_log = logging.getLogger('decoder.rru')
_log.setLevel(logging.DEBUG)

_RRU_BAUD = 19200
_RRU_PASSLEN = 12
_RRU_BEACONLEN = 17
_RRU_LOWBATT = 2.3  # Warn if battery voltage is below this many volts
_RRU_REFCHECK = 1800  # check ref after this many timeouts
_RRU_REFTHRESH = 0x2a30000  # 2 x 86400 x 256
_RRU_ENCODING = 'iso8859-1'
_RRU_MARKER = '_____127'  # trigger marker
_RRU_EOL = '\n'
# Serial port I/O read timeout in seconds
_RRU_IOTIMEOUT = 1.0
# List of handled responses from decoder
_REPLIES = [
    'ASCII',
    'CONFSET',
    'CONFGET',
    'INFOGET',
    'SITESURVEY',
    'TIMESTAMPGET',
    'EPOCHREFGET',
    'EPOCHREFSET',
    'EPOCHREFADJ1D',
    'PASSINGGET',
    'PASSINGINFOGET',
    'BEACONGET',
    'PREWARN',
]
# Internal Decoder configuration parameters,
# Some of these may be accessed through the
# sysconf/jsonconfig schema below.
# Values are sent to decoder in _sane().
_CONFINFO = {
    '01': {
        'label': 'Push Pre-Warn',
        'default': None,
        '00': 'disabled',
        '01': 'enabled'
    },
    '02': {
        'label': 'Blink/beep on repeated passing',
        'default': None,
        '00': 'disabled',
        '01': 'enabled'
    },
    '03': {
        'label': 'Impulse input or beep output',
        'default': None,
        '00': 'impulse-in',
        '01': 'beep-out'
    },
    '04': {
        'label': 'Auto-shutdown on power loss',
        # force disabled since USB cable connection often fails
        'default': '00',
        '00': 'disabled',
        '01': 'enabled'
    },
    '05': {
        'label': 'Operation Mode',
        # assume usb-timing is required unless explicity configured otherwise
        'default': '06',
        '00': 'active extension',
        '05': 'usb-kiosk',
        '06': 'usb-timing',
        '07': 'usb-store&copy'
    },
    '06': {
        'label': 'Channel ID',
        'default': None,
        '00': '1',
        '01': '2',
        '02': '3',
        '03': '4',
        '04': '5',
        '05': '6',
        '06': '7',
        '07': '8'
    },
    '07': {
        'label': 'Loop ID',
        'default': None,
        '00': '1',
        '01': '2',
        '02': '3',
        '03': '4',
        '04': '5',
        '05': '6',
        '06': '7',
        '07': '8'
    },
    '08': {
        'label': 'Loop Power',
        'default': None
    },
    '09': {
        'label': 'Blink dead-time',
        'default': None
    },
    '0a': {
        'label': 'Charging via USB',
        'default': None,
        '00': 'disabled',
        '01': 'enabled'
    },
    '0b': {
        'label': 'Use DTR',
        # always use DTR, unless configured otherwise
        'default': '01',
        '00': 'disabled',
        '01': 'enabled'
    },
    '0c': {
        'label': 'Alternate Channel Switching',
        'default': None,
        '00': 'disabled',
        '01': 'automatic',
        '02': 'force'
    },
    '0d': {
        'label': 'Box Mode',
        'default': None,
        '31': 'check',
        '32': 'deep-sleep',
        '33': 'health-check',
        '34': 'tracking',
        '41': 'usb-timing',
    },
    'a0': {
        'label': 'Tray Scan Power',
        'default': None,
    },
    'a1': {
        'label': 'Tray Scan interval',
        'default': None,
    },
    'a2': {
        'label': 'Tray Scan ramp up delay',
        'default': None,
    },
    'a3': {
        'label': 'Tray Scan row to column delay',
        'default': None,
    },
    'a4': {
        'label': 'Repeat Row and Column cycle for another N repetition',
        'default': None,
    },
    'b1': {
        'label': 'CheckSum',
        'default': None,
        '0': 'disabled',
        '1': 'enabled',
    },
    'b2': {
        'label': 'Push Passings',
        'default': None,
        '0': 'disabled',
        '1': 'enabled',
    }
}
# Labels for decoder information options
_INFOLBLS = {
    '01': 'Decoder ID',
    '02': 'Firmware Major Version',
    '03': 'Hardware Version',
    '04': 'Box Type',
    '05': 'Battery Voltage',
    '07': 'Battery State',
    '08': 'Battery Level',
    '09': 'Internal Temperature',
    '0a': 'Supply Voltage',
    '0b': 'Loop Status',
    '0c': 'Firmware Minor Version',
}
_BOX_TYPES = {
    '0a': 'active-ext',
    '1e': 'management-box',
    '28': 'usb-timing-box'
}
_BATTERY_STATES = {
    '00': 'Fault',
    '01': 'Charging',
    '02': 'Reduced Charging',
    '03': 'Discharging'
}
_LOOP_STATES = {
    '00': 'OK',
    '01': 'Fault',
    '02': 'Limit',
    '03': 'Overvoltage Error'
}

_CONFIG_SCHEMA = {
    'ttype': {
        'prompt': 'Race Result USB Active Decoder',
        'control': 'section'
    },
    'allowstored': {
        'type': 'bool',
        'subtext': 'Report stored passings?',
        'prompt': 'Allow Stored:',
        'hint': 'Stored passings will be reported as normal passings',
        'control': 'check',
        'default': True
    },
    'loopid': {
        'attr': '07',
        'prompt': 'Loop ID:',
        'hint': 'Active loop ID',
        'control': 'choice',
        'options': {
            '00': '1 (Base)',
            '01': '2',
            '02': '3',
            '03': '4',
            '04': '5',
            '05': '6',
            '06': '7',
            '07': '8'
        },
        'default': '00'
    },
    'looppower': {
        'attr': '08',
        'prompt': 'Loop Power:',
        'hint': 'Loop activation power',
        'type': 'float',
        'control': 'short',
        'subtext': '%',
        'default': 30.0,
    },
    'channel': {
        'attr': '06',
        'prompt': 'Channel ID:',
        'hint': 'Race Result wireless channel ID',
        'control': 'choice',
        'options': {
            '00': '1',
            '01': '2',
            '02': '3',
            '03': '4',
            '04': '5',
            '05': '6',
            '06': '7',
            '07': '8',
            'auto': 'Auto (Site Survey)',
        },
    },
    'hwopts': {
        'control': 'section',
        'prompt': 'Hardware Setup',
    },
    'impulse': {
        'prompt': 'Option Jack:',
        'control': 'choice',
        'attr': '03',
        'hint': 'Use the option jack for impulse in or beep output',
        'options': {
            '00': 'Impulse Input',
            '01': 'Beep Output'
        },
        'default': '00',
    },
    'mode': {
        'attr': '05',
        'control': 'choice',
        'prompt': 'Mode:',
        'hint': 'Specify extension timing mode',
        'default': '06',
        'options': {
            '00': 'Active Extension RS-232',
            '06': 'USB TIming Box',
            '05': 'USB Kiosk',
            '07': 'USB Store & Copy',
        },
    },
    'usedtr': {
        'attr': '0b',
        'type': 'bool',
        'subtext': 'Use DTR line?',
        'prompt': 'DTR Sync:',
        'hint': 'Use DTR signal to synchronise decoder with host computer',
        'control': 'check',
        'default': True
    },
    'charge': {
        'attr': '0a',
        'type': 'bool',
        'subtext': 'Charge via USB connection?',
        'prompt': 'Charge:',
        'hint': 'Charger decoder internal battery using USB power',
        'control': 'check',
        'default': True
    },
    'shutdown': {
        'attr': '04',
        'type': 'bool',
        'subtext': 'Shutdown on power loss?',
        'prompt': 'Shutdown:',
        'hint': 'Automatically shutdown decoder on power loss',
        'control': 'check',
        'default': False
    },
    'chanswitch': {
        'attr': '0c',
        'prompt': 'Channel Switch:',
        'hint': 'Alternate channel switching',
        'control': 'choice',
        'options': {
            '00': 'Disabled (V1)',
            '01': 'Automatic',
            '02': 'Forced (V2 Only)'
        },
    },
}


class rru(decoder):
    """Race Result USB Active Decoder"""

    def __init__(self):
        decoder.__init__(self)
        self._sitenoise = {
            1: 100,
            2: 100,
            3: 100,
            4: 100,
            5: 100,
            6: 100,
            7: 100,
            8: 100
        }
        self._config = {}
        self._boxname = None
        self._rrustamp = None
        self._rruht = None
        self._error = False
        self._io = None
        self._rdbuf = b''
        self._curreply = None  # current multi-line response mode
        self._lastpassing = 0
        self._lastrequest = None
        self._request_pending = 1
        self._allowstored = False
        self._autochannelid = None
        self._refcount = 0
        self._curport = None

    # API overrides
    def status(self):
        """Request status info from decoder."""
        for c in sorted(_CONFINFO):
            self.write(''.join(['CONFGET;', c]))
        for c in sorted(_INFOLBLS):
            self.write(''.join(['INFOGET;', c]))
        self.write('BEACONGET')
        self.write('PASSINGINFOGET')
        self.write('TIMESTAMPGET')

    def connected(self):
        """Return true if decoder is connected"""
        return self._boxname is not None and self._io is not None and self._request_pending < 10

    # Device-specific functions
    def _close(self):
        self._boxname = None
        if self._io is not None:
            _log.debug('Close connection')
            cp = self._io
            self._io = None
            cp.close()

    def _port(self, port):
        """Re-establish connection to supplied device port."""
        self._request_pending = 1
        self._rrustamp = None
        self._rruht = None
        self._close()
        if port is None and self._curport is not None:
            port = self._curport
        if port is None:
            _log.debug('Re-connect cancelled: port is None')
            return
        self._rdbuf = b''
        _log.debug('Connecting to %r', port)
        s = serial.Serial(baudrate=_RRU_BAUD,
                          rtscts=False,
                          timeout=_RRU_IOTIMEOUT)
        s.dtr = 0
        s.port = port
        s.open()
        self._io = s
        self._sane()
        self._cqueue.put(('_pendclr', ''))

    def _sane(self, data=None):
        """Load system config and then check decoder is properly configured"""
        sysconf.add_section('rru', _CONFIG_SCHEMA)
        self._allowstored = sysconf.get_value('rru', 'allowstored')
        _log.debug('Allow stored passings: %r', self._allowstored)
        setconfs = {}
        for opt in ('loopid', 'channel', 'impulse', 'chanswitch', 'mode'):
            nv = sysconf.get_value('rru', opt)
            if nv is not None:
                lbl = _CONFINFO[_CONFIG_SCHEMA[opt]['attr']]['label']
                setconfs[lbl] = nv
                _log.debug('Read value: %r = %r from sysconf', lbl, nv)
        for opt in ('usedtr', 'charge', 'shutdown'):
            if sysconf.has_value('rru', opt):
                nv = '00'
                if sysconf.get_value('rru', opt):
                    nv = '01'
                lbl = _CONFINFO[_CONFIG_SCHEMA[opt]['attr']]['label']
                setconfs[lbl] = nv
                _log.debug('Read value: %r = %r from sysconf', lbl, nv)
        if sysconf.has_value('rru', 'looppower'):
            # WTF Race Result? 0x0-0x64
            nv = round(sysconf.get_value('rru', 'looppower'))
            nv = '%02x' % ((min(100, max(0, nv))))
            lbl = _CONFINFO[_CONFIG_SCHEMA['looppower']['attr']]['label']
            setconfs[lbl] = nv
            _log.debug('Read value: %r = %r from sysconf', lbl, nv)

        # re-build internal configuration
        self._config = {}
        for opt in _CONFINFO:
            ko = _CONFINFO[opt]
            kn = ko['label']
            if kn in setconfs and setconfs[kn] is not None:
                self._config[opt] = setconfs[kn]
                sv = setconfs[kn]
                if sv in ko:
                    sv = ko[sv]
                _log.debug('Set %s: %s', kn, sv)
            elif ko['default'] is not None:
                self._config[opt] = ko['default']

        # if channel id set already by site survey, keep it
        if self._autochannelid is not None:
            self._config['06'] = self._autochannelid

        # Clear any messages in queue
        self._dump_queue()

        # Set protocol
        self.write('ASCII')
        # Fetch decoder ID
        self.write('INFOGET;01')
        # Set options, ordering ensures Box mode is set before impulse in
        for opt in [
                '05', '01', '02', '04', '07', '08', '09', '0a', '0b', '0c',
                '0d', 'a0', 'a1', 'a2', 'a3', 'a4', 'b1', 'b2', '03'
        ]:
            if opt in self._config and self._config[opt] is not None:
                self.write(';'.join(['CONFSET', opt, self._config[opt]]))
        # Check if site survey is required
        if '06' in self._config and self._config['06'] is not None:
            if self._config['06'].lower() == 'auto':
                _log.info('Performing site survey to set Channel ID')
                self.write('SITESURVEY')
            else:
                self.write(';'.join(['CONFSET', '06', self._config['06']]))
        # Request current epoch ref setting
        self.write('EPOCHREFGET')
        # Request current number of ticks
        self.write('TIMESTAMPGET')
        # Request current memory status
        self.write('PASSINGINFOGET')

    def _sync(self, data=None):
        _log.debug('Performing blocking DTR sync')
        # for the purpose of sync, the "epoch" is considered to be
        # midnight localtime of the current day
        self._rrustamp = None
        self._rruht = None
        # Determine the 'reference' epoch
        nt = tod.now()
        ntt = nt.truncate(0)
        ntd = (nt - ntt).timeval
        if ntd < 0.1 or ntd > 0.9:
            _log.debug('Sleeping 0.3s')
            sleep(0.3)
            nt = tod.now()
        ntt = nt.truncate(0)
        ett = ntt + tod.ONE
        _log.debug('Host reference time: %s', ett.rawtime())
        es = 'EPOCHREFSET;{0:08x}'.format(int(ett.timeval))
        self._write(es)
        _log.debug('Waiting for top of second')
        acceptval = tod.tod('0.001')
        diff = ett - nt
        while diff > acceptval and diff < tod.ONE:
            sleep(0.0005)
            nt = tod.now()
            diff = ett - nt
        _log.debug('Set DTR')
        self._io.dtr = 1
        sleep(0.2)
        _log.debug('Clear DTR')
        self._io.dtr = 0
        self._cqueue.put(('_pendclr', ''))

    def _start_session(self, data=None):
        """Reset decoder to start a new session"""
        # On USB decoder, session is always active, this
        # method performs a hard reset/clear/sync
        self._clear(data)
        _log.info('Start session')

    def _stop_session(self, data=None):
        """Stop the current timing session"""
        # USB decoder will continue collecting passings, but
        # they are not received by handler - the passings can
        # be fetched by restarting connection without reset
        self._rrustamp = None
        _log.info('Stop session')

    def _clear(self, data=None):
        _log.info('Requesting new session')
        self._rrustamp = None
        self._boxname = None
        self._write('RESET')
        while True:
            m = self._readline()
            #_log.debug('RECV: %r', m)
            if m == 'AUTOBOOT':
                sleep(1.0)
                break
        self._lastrequest = 0
        self._lastpassing = 0
        self._sane()

    def _dump_queue(self):
        try:
            while not self._cqueue.empty():
                m = self._cqueue.get_nowait()
                _log.debug('Dumping command: %r', m)
                self._cqueue.task_done()
        except Exception as e:
            _log.debug('$s emptying queue: %s', e.__class__.__name__, e)

    def _write(self, msg):
        if self._io is not None:
            ob = (msg + _RRU_EOL)
            self._io.write(ob.encode(_RRU_ENCODING))
            #_log.debug('SEND: %r', ob)

    def _pendclr(self, msg):
        _log.debug('Enable passing request')
        self._request_pending = 0

    def _tstotod(self, ts):
        """Convert a race result timestamp to time of day."""
        ret = None
        try:
            ti = int(ts, 16) - self._rrustamp
            tsec = decimal.Decimal(ti // 256) + decimal.Decimal(ti % 256) / 256
            nsec = (self._rruht.timeval + tsec) % 86400
            if nsec < 0:
                _log.debug('Negative timestamp: %r', nsec)
                nsec = 86400 + nsec
            ret = tod.tod(nsec)
        except Exception as e:
            _log.error('%s converting timeval %r: %s', e.__class__.__name__,
                       ts, e)
        return ret

    def _confmsg(self, cid, val):
        """Handle config response."""
        lbl = cid
        vbl = val
        if cid in _CONFINFO:
            option = _CONFINFO[cid]
            if 'label' in option:
                lbl = option['label']
            if cid == '08':
                vbl = '{0:d}%'.format(int(val, 16))
            elif cid == '07':
                vbl = '{0:d}'.format(int(val, 16) + 1)
            else:
                if val in option:
                    vbl = option[val]
        if cid in self._config:
            # check decoder has the desired value
            if val != self._config[cid] and self._config[cid] != 'auto':
                if self._curreply == 'CONFSET':
                    _log.error('Error setting config %s, desired:%r actual:%r',
                               lbl, self._config[cid], val)
                else:
                    _log.info('Updating config %s: %s => %s', lbl, val,
                              self._config[cid])
                    self.write(';'.join(['CONFSET', cid, self._config[cid]]))
            else:
                _log.info('Config %s: %s', lbl, vbl)
        else:
            if cid in ['07', '08']:
                _log.info('Config %s: %s', lbl, vbl)
            else:
                _log.debug('Config %s: %s', lbl, vbl)

    def _infomsg(self, pid, val):
        """Show and save decoder info message."""
        lbl = pid
        vbl = val
        if pid in _INFOLBLS:
            lbl = _INFOLBLS[pid]
            if pid == '01':  # ID
                vbl = 'A-{0:d}'.format(int(val, 16))
                self._boxname = vbl
            elif pid == '02':  # Firmware
                vbl = 'v{0:0.1f}'.format(int(val, 16) / 10)
            elif pid == '03':  # Hardware
                vbl = 'v{0:0.1f}'.format(int(val, 16) / 10)
            elif pid == '04':  # Box Type
                if val in _BOX_TYPES:
                    vbl = _BOX_TYPES[val]
            elif pid == '05':  # Batt Voltage
                vbl = '{0:0.1f}V'.format(int(val, 16) / 10)
            elif pid == '07':  # Battery State
                if val in _BATTERY_STATES:
                    vbl = _BATTERY_STATES[val]
            elif pid == '08':  # Battery Level
                vbl = '{0:d}%'.format(int(val, 16))
            elif pid == '09':  # Int Temp
                vbl = '{0:d}\xb0C'.format(int(val, 16))
            elif pid == '0a':  # Supply Voltage
                vbl = '{0:0.1f}V'.format(int(val, 16) / 10)
            elif pid == '0b':  # Loop Status
                if val in _LOOP_STATES:
                    vbl = _LOOP_STATES[val]
            _log.info('Info %s: %s', lbl, vbl)
        else:
            _log.info('Info [undocumented] %s: %s', lbl, vbl)

    def _refgetmsg(self, epoch, stime):
        """Collect the epoch ref and system tick message."""
        self._rruht = tod.mkagg(int(epoch, 16))
        self._rrustamp = int(stime, 16)
        _log.debug('Reference ticks: %r @ %r', self._rrustamp,
                   self._rruht.rawtime())
        if self._rrustamp == 0:
            _log.info('Decoder time is not set, requesting sync')
            self.sync()

    def _timestampchk(self, ticks):
        """Receive the number of ticks on the decoder."""
        tcnt = int(ticks, 16)
        _log.debug('Box tick count: %r', tcnt)
        if tcnt > _RRU_REFTHRESH:
            _log.info('Box tick threshold exceeded, adjusting ref')
            self.write('EPOCHREFADJ1D')

    def _passinginfomsg(self, mv):
        """Receive info about internal passing memory."""
        if len(mv) == 5:
            pcount = int(mv[0], 16)
            if pcount > 0:
                pfirst = int(mv[1], 16)
                pftime = self._tstotod(mv[2])
                plast = int(mv[3], 16)
                pltime = self._tstotod(mv[4])
                _log.info('Info %r Passings, %r@%s - %r@%s', pcount, pfirst,
                          pftime.rawtime(2), plast, pltime.rawtime(2))
                if self._lastrequest is None:
                    self._lastpassing = plast + 1
                else:
                    if plast + 1 < self._lastpassing:
                        _log.warning(
                            'Decoder memory/ID mismatch last=%r, req=%r, clear/sync required.',
                            plast, self._lastpassing)
                        self._lastpassing = plast + 1
            else:
                _log.info('Info No Passings')
        else:
            _log.debug('Non-passinginfo message: %r', mv)

    def _passingmsg(self, mv):
        """Receive a passing from the decoder."""
        if len(mv) == _RRU_PASSLEN:
            # USB decoder doesn't return passing ID, use internal count
            istr = str(self._lastpassing)
            tagid = mv[0]  # [TranspCode:string]
            wuc = mv[1]  # [WakeupCounter:4]
            timestr = mv[2]  # [Time:8]
            hits = mv[3]  # [Hits:2]
            rssi = mv[4]  # [RSSI:2]
            battery = mv[5]  # [Battery:2]
            loopid = mv[8]  # [LoopId:1]
            adata = mv[10]  # [InternalActiveData:2]

            # Check values
            if not loopid:
                loopid = 'C1'  # add faked id for passives
            else:
                loopid = strops.id2chan(int(loopid, 16) + 1)
            activestore = False
            if adata:
                aval = int(adata, 16)
                activestore = (int(adata, 16) & 0x40) == 0x40
                if activestore:
                    _log.debug('Stored passing %s: %r', tagid, adata)
            if tagid == _RRU_MARKER:
                tagid = ''

            if battery and tagid:
                try:
                    bv = int(battery, 16) / 10
                    if bv < _RRU_LOWBATT:
                        _log.warning('Low battery %s: %0.1fV', tagid, bv)
                except Exception as e:
                    _log.debug('%s reading battery voltage: %s',
                               e.__class__.__name__, e)

            if hits and rssi and tagid:
                try:
                    hitcount = int(hits, 16)
                    rssival = int(rssi, 16)
                    twofour = -90 + ((rssival & 0x70) >> 2)
                    lstrength = 1 + (rssival & 0x0f)
                    if lstrength < 3 or twofour < -82 or hitcount < 3:
                        _log.warning(
                            'Poor read %s: Hits:%d RSSI:%ddBm Loop:%ddB',
                            tagid, hitcount, twofour, lstrength)
                except Exception as e:
                    _log.debug('%s reading hits/RSSI: %s',
                               e.__class__.__name__, e)

            # emit a decoder log line TBD
            _log.log(DECODER_LOG_LEVEL, ';'.join(mv))

            # accept valid passings and trigger callback
            t = self._tstotod(timestr)
            if t is not None:
                t.index = istr
                t.chan = loopid
                t.refid = tagid
                t.source = self._boxname
                if not activestore or self._allowstored:
                    self._trig(t)
                else:
                    pass
            self._lastpassing += 1
        elif len(mv) == 2:
            resp = int(mv[0], 16)
            rcount = int(mv[1], 16)
            if resp != self._lastrequest:
                _log.debug('Sequence mismatch request: %r, response: %r',
                           self._lastrequest, resp)
                self._lastpassing = 0
            elif rcount > 0:
                _log.debug('Receiving %r passings', rcount)
        else:
            _log.debug('Non-passing message: %r', mv)

    def _beaconmsg(self, mv):
        """Receive a beacon from the decoder."""
        if len(mv) == _RRU_BEACONLEN:
            # noise/transponder averages
            chid = int(mv[5], 16) + 1
            chnoise = 10.0 * int(mv[12], 16)
            tlqi = int(mv[13], 16) / 2.56
            trssi = -90 + int(mv[14], 16)
            _log.info('Info Ch {0} Noise: {1:0.0f}%'.format(chid, chnoise))
            _log.info('Info Avg LQI: {0:0.0f}%'.format(tlqi))
            _log.info('Info Avg RSSI: {}dBm'.format(trssi))
        elif len(mv) == 1:
            bcount = int(mv[0], 16)
            _log.debug('Receiving %r beacons', bcount)
        else:
            _log.debug('Non-beacon message: %r', mv)

    def _idupdate(self, reqid, minid):
        """Handle an empty PASSINGGET response."""
        resp = int(reqid, 16)
        newid = int(minid, 16)
        if resp != self._lastrequest:
            _log.error(
                'ID mismatch: request=%r, response=%r:%r, reset to first passing',
                self._lastrequest, resp, newid)
            self._lastpassing = newid
        else:
            if resp < newid:
                _log.warning('Data loss: %r passings not received',
                             newid - resp)
                self._lastpassing = newid
            # otherwise no missed passings, ignore error

    def _surveymsg(self, chan, noise):
        """Receive a site survey update."""
        channo = int(chan, 16) + 1
        if channo in self._sitenoise:
            self._sitenoise[channo] = 10 * int(noise, 16)
        else:
            _log.debug('Unknown channel in site survey: %r', channo)

    def _chansurf(self):
        """Examine survey for a better channel and hop if needed."""
        ch = None
        cv = 55  # don't accept any channel with noise over 50%
        lv = []
        for c in sorted(self._sitenoise, key=strops.rand_key):
            nv = self._sitenoise[c]
            lv.append('{}:{:d}%'.format(c, nv))
            if nv < cv:
                ch = c
                cv = nv
        _log.debug('Site survey: %s', ' '.join(lv))
        if ch is not None:
            _log.info('Auto-selected channel %r, Noise: %d%%', ch, cv)
            sv = '{0:02x}'.format(ch - 1)
            self._autochannelid = sv
            m = 'CONFSET;06;{}'.format(sv)
            self.write(m)
        else:
            _log.warning('Unable to find a suitable channel')

    def _handlereply(self, mv):
        """Process the body of a decoder response."""
        if self._curreply == 'PASSINGGET':
            self._passingmsg(mv)
        elif self._curreply == 'PASSINGINFOGET':
            self._passinginfomsg(mv)
        elif self._curreply == 'PASSINGIDERROR':
            if len(mv) == 2:
                self._idupdate(mv[0], mv[1])
        elif self._curreply == 'INFOGET':
            if len(mv) == 2:
                self._infomsg(mv[0], mv[1])
        elif self._curreply in ['EPOCHREFGET', 'EPOCHREFSET', 'EPOCHREFADJ1D']:
            if len(mv) == 2:
                self._refgetmsg(mv[0], mv[1])
        elif self._curreply in ['CONFGET', 'CONFSET']:
            if len(mv) == 2:
                self._confmsg(mv[0], mv[1])
        elif self._curreply == 'TIMESTAMPGET':
            if len(mv) == 1:
                self._timestampchk(mv[0])
        elif self._curreply == 'SITESURVEY':
            if len(mv) == 2:
                self._surveymsg(mv[0], mv[1])
        elif self._curreply == 'BEACONGET':
            self._beaconmsg(mv)
        else:
            _log.debug('%r : %r', self._curreply, mv)

    def _procline(self, l):
        """Handle the next line of response from decoder."""
        mv = l.split(';')
        if len(mv) > 0:
            if mv[0] == '#P':
                # pushed passing overrides cmd/reply logic
                self._passingmsg(mv[1:])
            elif mv[0] == 'PREWARN':
                self._curreply = mv[0]
            elif mv[0] == 'EOR':
                if self._curreply in ['PASSINGGET', 'PASSINGIDERROR']:
                    self._request_pending = 0
                if self._curreply == 'SITESURVEY':
                    self._chansurf()
                self._curreply = None
            elif mv[0] in _REPLIES:
                if self._curreply is not None:
                    _log.debug('Protocol error: %r not terminated',
                               self._curreply)
                self._curreply = mv[0]
                if mv[1] != '00':
                    if self._curreply == 'PASSINGGET' and mv[1] == '10':
                        self._curreply = 'PASSINGIDERROR'
                    else:
                        _log.debug('%r error: %r', self._curreply, mv[1])
            elif mv[0] == 'AUTOBOOT':
                _log.warning('Re-connecting restarted decoder')
                self._request_pending = 1
                self._sane()
            elif 'rrActive' in mv[0]:
                _log.warning('Decoder reset - waiting for it to start')
                self._request_pending = 1
                sleep(3.0)
                self._dump_queue()
            else:
                if self._curreply is not None:
                    self._handlereply(mv)

    def _readline(self):
        """Read from the decoder until end of line or timeout condition."""
        ret = None
        ch = self._io.read(1)
        while ch != b'':
            if ch == b'\n':
                if len(self._rdbuf) > 0:
                    # linefeed ends the current 'message'
                    ret = self._rdbuf.lstrip(b'\0').decode(_RRU_ENCODING)
                    self._rdbuf = b''
                else:
                    ret = 'EOR'  # Flag end of response
                break
            else:
                self._rdbuf += ch
            ch = self._io.read(1)
        return ret

    def _request_next(self):
        """Queue a passingget request if the reftime is set."""
        if self._rrustamp is not None:
            if not self._request_pending:
                self._request_pending = 1
                es = 'PASSINGGET;{0:08x}'.format(self._lastpassing)
                self._lastrequest = self._lastpassing
                self.write(es)
                self._refcount += 1
                if self._refcount > _RRU_REFCHECK:
                    self.write('TIMESTAMPGET')
                    self._refcount = 0
            else:
                self._request_pending += 1

    def run(self):
        """Decoder main loop."""
        _log.debug('Starting')
        self._running = True
        while self._running:
            try:
                m = None  # next commmand
                if self._io is not None:
                    # Fetch all responses from unit
                    refetch = False
                    while True:
                        l = self._readline()
                        if l is None:
                            # wait for sitesurvey
                            if self._curreply != 'SITESURVEY':
                                # on time out, request passings
                                refetch = True
                                break
                        else:
                            #_log.debug('RECV: %r', l)
                            self._procline(l)
                            if self._curreply == 'PREWARN':
                                # Note: this does not work
                                refetch = True
                            if l == 'EOR':
                                break
                    if refetch:
                        self._request_next()
                    m = self._cqueue.get_nowait()
                else:
                    m = self._cqueue.get()
                self._cqueue.task_done()
                self._proccmd(m)
            except queue.Empty:
                pass
            except serial.SerialException as e:
                self._close()
                _log.error('%s, decoder disconnected', e.__class__.__name__)
            except Exception as e:
                _log.critical('%s: %s', e.__class__.__name__, e)
                self._running = False
                self._boxname = None
        self.setcb()
        _log.debug('Exiting')



# SPDX-License-Identifier: MIT
"""UNT4 Packet Wrapper."""

# UNT4 mode 1 constants
NUL = b'\x00'
SOH = b'\x01'
STX = b'\x02'
EOT = b'\x04'
CR = b'\x0d'
LF = b'\x0a'
ERL = b'\x0b'
ERP = b'\x0c'
DLE = b'\x10'
DC2 = b'\x12'
DC3 = b'\x13'
DC4 = b'\x14'
US = b'\x1f'

ENCMAP = {
    chr(NUL[0]): '<0>',
    chr(SOH[0]): '<O>',
    chr(STX[0]): '<T>',
    chr(EOT[0]): '<E>',
    chr(CR[0]): '<R>',
    chr(LF[0]): '<A>',
    chr(ERL[0]): '<L>',
    chr(ERP[0]): '<P>',
    chr(DLE[0]): '<D>',
    chr(DC2[0]): '<2>',
    chr(DC3[0]): '<3>',
    chr(DC4[0]): '<4>',
    chr(US[0]): '<U>'
}

TRANSMAP = str.maketrans({
    SOH[0]: 0x20,
    STX[0]: 0x20,
    EOT[0]: 0x20,
    ERL[0]: 0x20,
    ERP[0]: 0x20,
    DLE[0]: 0x20,
    DC2[0]: 0x20,
    DC3[0]: 0x20,
    DC4[0]: 0x20,
})


def encode(ubuf=''):
    """Encode the unt4 buffer for use over telegraph."""
    # escape special char
    ubuf = ubuf.replace('<', '<>')
    # escape control chars
    for key in ENCMAP:
        ubuf = ubuf.replace(key, ENCMAP[key])
    return ubuf


def decode(tbuf=''):
    """Decode the telegraph buffer to unt4 pack."""
    # decode control chars
    for key in ENCMAP:
        tbuf = tbuf.replace(ENCMAP[key], key)
    # decode special char
    tbuf = tbuf.replace('<>', '<')
    return tbuf


# UNT4 Packet class
class unt4:
    """UNT4 Packet Class."""

    def __init__(self,
                 unt4str=None,
                 prefix=None,
                 header='',
                 erp=False,
                 erl=False,
                 xx=None,
                 yy=None,
                 text=''):
        self.prefix = prefix  # <DC2>, <DC3>, etc
        self.header = header.translate(TRANSMAP)
        self.erp = erp  # true for general clearing <ERP>
        self.erl = erl  # true for <ERL>
        self.xx = xx  # input column 0-99
        self.yy = yy  # input row 0-99
        self.text = text.translate(TRANSMAP)
        if unt4str is not None:
            self.unpack(unt4str)

    def unpack(self, unt4str=''):
        """Unpack the UNT4 unicode string into this object."""
        if len(unt4str) > 2 and ord(unt4str[0]) == SOH[0] \
                            and ord(unt4str[-1]) == EOT[0]:
            self.prefix = None
            newhead = ''
            newtext = ''
            self.erl = False
            self.erp = False
            head = True  # All text before STX is considered header
            stx = False
            dle = False
            dlebuf = ''
            i = 1
            packlen = len(unt4str) - 1
            while i < packlen:
                och = ord(unt4str[i])
                if och == STX[0]:
                    stx = True
                    head = False
                elif och == DLE[0] and stx:
                    dle = True
                elif dle:
                    dlebuf += unt4str[i]
                    if len(dlebuf) == 4:
                        dle = False
                elif head:
                    if och in (DC2[0], DC3[0], DC4[0]):
                        self.prefix = och  # assume pfx before head text
                    else:
                        newhead += unt4str[i]
                elif stx:
                    if och == ERL[0]:
                        self.erl = True
                    elif och == ERP[0]:
                        self.erp = True
                    else:
                        newtext += unt4str[i]
                i += 1
            if len(dlebuf) == 4:
                self.xx = int(dlebuf[:2])
                self.yy = int(dlebuf[2:])
            self.header = newhead
            self.text = newtext

    def pack(self):
        """Return Omega Style UNT4 unicode string packet."""
        head = ''
        text = ''
        if self.erp:  # overrides any other message content
            text = chr(STX[0]) + chr(ERP[0])
        else:
            head = self.header
            if self.prefix is not None:
                head = chr(self.prefix) + head
            if self.xx is not None and self.yy is not None:
                text += chr(DLE[0]) + '{0:02d}{1:02d}'.format(
                    self.xx % 100, self.yy % 100)
            if self.text:
                text += self.text
            if self.erl:
                text += chr(ERL[0])
            if len(text) > 0:
                text = chr(STX[0]) + text
        return chr(SOH[0]) + head + text + chr(EOT[0])


# Pre-defined messages
GENERAL_CLEARING = unt4(erp=True)
GENERAL_EMPTY = unt4(xx=0, yy=0, text='')



# SPDX-License-Identifier: MIT
"""Telegraph

 Telegraph is a thin wrapper over paho.mqtt.client, which provides
 a MQTT pub/sub interface, altered for use with metarace applications.

 Example:

	import metarace
	from metarace import telegraph
	metarace.init()
	
	def messagecb(topic, message):
	    obj = telegraph.from_json(message)
	    ...
	
	t = telegraph.telegraph()
	t.set_will_json({'example':[]}, 'thetopic')
	t.subscribe('thetopic')
	t.setcb(messagecb)
	t.start()
	...
	t.publish_json({'example':[1,2,3]}, 'thetopic')

 Message callback functions receive two named parameters 'topic' and
 'message' which are both unicode strings. The message callback is run in
 the telegraph thread context. Use the convenience function "from_json"
 to convert a message from json into a python object. See defcallback
 for an example.

 Configuration is via metarace system config (metarace.json), under
 section 'telegraph':

  key: (type) Description [default]
  --
  host : (string) MQTT broker, None to disable ['localhost']
  port: (int) MQTT port [1883/8883]
  username : (string) username [None]
  password : (string) password [None]
  usetls : (bool) if True, connect to server over TLS [False]
  persist : (bool) if true, open a persistent connection to broker [False]
  clientid : (string) provide an explicit client id [None]
  qos : (int) default QOS to use for subscribe and publish [0]
  debug : (bool) if True, enable logging in MQTT library [False]
  deftopic : (string) a default publish topic [None]


"""

import threading
import queue
import logging
import json
import paho.mqtt.client as mqtt
from paho.mqtt.enums import CallbackAPIVersion
from uuid import uuid4
import metarace

_QUEUE_TIMEOUT = 2

# module logger
_log = logging.getLogger('telegraph')
_log.setLevel(logging.DEBUG)

_CONFIG_SCHEMA = {
    'ttype': {
        'prompt': 'MQTT Broker Details',
        'control': 'section'
    },
    'host': {
        'prompt': 'Host:',
        'hint': 'Hostname or IP of MQTT broker',
        'default': 'localhost',
    },
    'port': {
        'prompt': 'Port:',
        'control': 'short',
        'hint': 'TCP port number of MQTT broker',
        'type': 'int',
    },
    'usetls': {
        'prompt': 'Security:',
        'subtext': 'Use TLS?',
        'type': 'bool',
        'control': 'check',
        'hint': 'Connect to MQTT broker using TLS',
        'default': False
    },
    'username': {
        'prompt': 'Username:',
        'hint': 'Username on MQTT broker if required'
    },
    'password': {
        'prompt': 'Password:',
        'hint': 'Password on MQTT broker if required'
    },
    'debug': {
        'prompt': 'Log:',
        'control': 'check',
        'type': 'bool',
        'subtext': 'Debug MQTT connection?',
        'hint': 'Log detailed information on connection to MQTT broken',
        'default': False,
    },
    'ssec': {
        'prompt': 'Session Options',
        'control': 'section',
    },
    'clientid': {
        'prompt': 'Client ID:',
        'hint': 'Specify client id for MQTT connection'
    },
    'persist': {
        'prompt': 'Session:',
        'subtext': 'Persistent?',
        'type': 'bool',
        'control': 'check',
        'hint': 'Request persistent session on broker',
        'default': True,
    },
    'qos': {
        'prompt': 'QoS:',
        'control': 'choice',
        'hint': 'Default message QOS',
        'type': 'int',
        'options': {
            '0': '0 - At most once',
            '1': '1 - At least once',
            '2': '2 - Exactly once'
        },
        'default': 0,
    },
    'deftopic': {
        'prompt': 'Default Topic:',
        'hint': 'Default topic for messages published without one',
    },
}


def from_json(payload=None):
    """Return message payload decoded from json, or None."""
    ret = None
    try:
        ret = json.loads(payload)
    except Exception as e:
        _log.warning('%s decoding JSON payload: %s', e.__class__.__name__, e)
    return ret


def defcallback(topic=None, message=None):
    """Default message receive callback function."""
    ob = from_json(message)
    if ob is not None:
        _log.debug('RCV %r: %r', topic, ob)
    else:
        _log.debug('RCV %r: %r', topic, message)


class telegraph(threading.Thread):
    """Metarace telegraph server thread."""

    def subscribe(self, topic=None, qos=None):
        """Add topic to the set of subscriptions with optional qos."""
        if topic:
            self.__subscriptions[topic] = qos
            if self.__connected:
                self.__queue.put_nowait(('SUBSCRIBE', topic, qos))

    def unsubscribe(self, topic=None):
        """Remove topic from the set of subscriptions."""
        if topic and topic in self.__subscriptions:
            del self.__subscriptions[topic]
            if self.__connected:
                self.__queue.put_nowait(('UNSUBSCRIBE', topic))

    def unsubscribe_all(self):
        """Remove all topics from the set of subscriptions."""
        topics = self.__subscriptions
        self.__subscriptions = {}
        for topic in topics:
            if self.__connected:
                self.__queue.put_nowait(('UNSUBSCRIBE', topic))
        topics = None

    def setcb(self, func=None):
        """Set the message receive callback function."""
        if func is not None:
            self.__cb = func
        else:
            self.__cb = defcallback

    def set_deftopic(self, topic=None):
        """Set or clear the default publish topic."""
        if isinstance(topic, str) and topic:
            self.__deftopic = topic
        else:
            self.__deftopic = None
        _log.debug('Default publish topic set to: %r', self.__deftopic)

    def set_will_json(self, obj=None, topic=None, qos=None, retain=False):
        """Pack the provided object into JSON and set as will."""
        try:
            self.set_will(json.dumps(obj), topic, qos, retain)
        except Exception as e:
            _log.error('Error setting will object %r: %s', obj, e)

    def set_will(self, message=None, topic=None, qos=None, retain=False):
        """Set or clear the last will with the broker."""
        if not self.__connect_pending and not self.__connected:
            if topic is not None:
                nqos = qos
                if nqos is None:
                    nqos = self.__qos
                payload = None
                if message is not None:
                    payload = message.encode('utf-8')
                self.__client.will_set(topic, payload, nqos, retain)
                _log.debug('Will set on topic %r', topic)
            else:
                self.__client.will_clear()
                _log.debug('Cleared will')
        else:
            _log.error('Unable to set will, already connected')

    def connected(self):
        """Return true if connected."""
        return self.__connected

    def unsent(self):
        """Return number of queued messages remaining unsent."""
        # _log.debug('Unsent messages: %r', self.__unsent)
        return len(self.__unsent)

    def reconnect(self):
        """Request re-connection to broker."""
        self.__queue.put_nowait(('RECONNECT', True))

    def exit(self, msg=None):
        """Request thread termination."""
        self.__running = False
        self.__queue.put_nowait(('EXIT', msg))

    def wait(self):
        """Suspend calling thread until command queue is processed."""
        self.__queue.join()

    def publish(self,
                message=None,
                topic=None,
                qos=None,
                retain=False,
                timeout=None):
        """Queue provided message for publishing to nominated topic.

        If timeout (float seconds > 0) is provided, telegraph
        will wait up to timeout seconds after publishing for
        completion before re-processing the command queue.

        For QoS == 0, timeout applies to message delivery
        For QoS == 1, timeout applies to PUBACK
        For QoS == 2, timeout applies to PUBCOMP

        This method returns immediately even if timeout set.
        """
        self.__queue.put_nowait(
            ('PUBLISH', topic, message, qos, retain, timeout))

    def publish_json(self,
                     obj=None,
                     topic=None,
                     qos=None,
                     retain=False,
                     cls=None,
                     indent=None,
                     timeout=None):
        """Pack the provided object into JSON and publish to topic."""
        try:
            self.publish(json.dumps(obj, cls=cls, indent=indent), topic, qos,
                         retain, timeout)
        except Exception as e:
            _log.error('Error publishing object %r: %s', obj, e)

    def __init__(self):
        """Constructor."""
        threading.Thread.__init__(self, daemon=True)
        self.__queue = queue.Queue()
        self.__cb = defcallback
        self.__subscriptions = {}
        self.__connected = False
        self.__connect_pending = False
        self.__cid = None
        self.__resub = True
        self.__doreconnect = False
        self.__unsent = set()
        self.__publishLock = threading.Lock()

        metarace.sysconf.add_section('telegraph', _CONFIG_SCHEMA)
        self.__host = metarace.sysconf.get_value('telegraph', 'host')
        self.__port = metarace.sysconf.get_value('telegraph', 'port')
        if self.__port is None:
            self.__port = 1883
        self.__deftopic = metarace.sysconf.get_value('telegraph', 'deftopic')
        self.__qos = metarace.sysconf.get_value('telegraph', 'qos')
        self.__persist = metarace.sysconf.get_value('telegraph', 'persist')
        self.__cid = metarace.sysconf.get_value('telegraph', 'clientid')

        # check values
        if self.__qos > 2:
            _log.info('Invalid QOS %r set to %r', self.__qos, 2)
            self.__qos = 2
        if not self.__cid:
            self.__cid = str(uuid4())
        _log.debug('Using QoS: %r', self.__qos)
        _log.debug('Using client id: %r', self.__cid)
        _log.debug('Persistent connection: %r', self.__persist)

        # create mqtt client
        self.__client = mqtt.Client(
            callback_api_version=CallbackAPIVersion.VERSION2,
            client_id=self.__cid,
            clean_session=not self.__persist)

        if metarace.sysconf.get_value('telegraph', 'debug'):
            _log.debug('Enabling mqtt/paho debug')
            mqlog = logging.getLogger('telegraph.mqtt')
            mqlog.setLevel(logging.DEBUG)
            self.__client.enable_logger(mqlog)

        if metarace.sysconf.get_value('telegraph', 'usetls'):
            _log.debug('Enabling TLS connection')
            if not metarace.sysconf.has_value('telegraph', 'port'):
                # Update port for TLS if not exlicitly set
                self.__port = 8883
                _log.debug('Set port to %r', self.__port)
            self.__client.tls_set()
        username = metarace.sysconf.get_value('telegraph', 'username')
        password = metarace.sysconf.get_value('telegraph', 'password')
        if username and password:
            self.__client.username_pw_set(username, password)

        self.__client.reconnect_delay_set(2, 16)
        self.__client.on_message = self.__on_message
        self.__client.on_connect = self.__on_connect
        self.__client.on_disconnect = self.__on_disconnect
        self.__client.on_publish = self.__on_publish
        if self.__host:
            self.__doreconnect = True
        self.__running = False

    def __reconnect(self):
        if not self.__connect_pending:
            if self.__connected:
                _log.debug('Disconnecting client')
                self.__client.disconnect()
                self.__client.loop_stop()
            if self.__host:
                _log.debug('Connecting to %s:%d', self.__host, self.__port)
                self.__connect_pending = True
                self.__client.connect_async(self.__host, self.__port)
                self.__client.loop_start()

    # PAHO methods - Callback API Version=2
    def __on_connect(self, client, userdata, flags, reason_code, properties):
        if reason_code == 0:
            _log.debug('Connect %r: %r/%r', client._client_id, flags,
                       reason_code)
            if not self.__resub and self.__persist and flags.session_present:
                _log.debug('Resumed existing session for %r',
                           client._client_id)
            else:
                _log.debug('Assuming Clean session for %r', client._client_id)
                s = []
                for t in self.__subscriptions:
                    nqos = self.__subscriptions[t]
                    if nqos is None:
                        nqos = self.__qos
                    s.append((t, nqos))
                if len(s) > 0:
                    _log.debug('Subscribe topics: %r', s)
                    self.__client.subscribe(s)
                self.__resub = False
            self.__connected = True
        else:
            _log.info('Connect failed with error %r: %r', reason_code,
                      mqtt.connack_string(reason_code))
            self.__connected = False
        self.__connect_pending = False

    def __on_disconnect(self, client, userdata, flags, reason_code,
                        properties):
        _log.debug('Disconnect %r: %r', client._client_id, reason_code)
        self.__connected = False
        # Note: PAHO lib will attempt re-connection automatically

    def __on_message(self, client, userdata, message):
        #_log.debug('Message from %r: %r', client._client_id, message)
        self.__cb(topic=message.topic, message=message.payload.decode('utf-8'))

    def __on_publish(self, client, userdata, mid, reason_code, properties):
        with self.__publishLock:
            try:
                self.__unsent.remove(mid)
            except Exception as e:
                _log.warning('%s removing published MID=%d: %s',
                             e.__class__.__name__, mid, e)

    def run(self):
        """Called via threading.Thread.start()."""
        self.__running = True
        if self.__host:
            _log.debug('Starting')
        else:
            _log.debug('Not connected')
        while self.__running:
            try:
                # Check connection status
                if self.__host and self.__doreconnect:
                    self.__doreconnect = False
                    if not self.__connect_pending:
                        self.__reconnect()
                # Process command queue
                while self.__running:
                    m = self.__queue.get(timeout=_QUEUE_TIMEOUT)
                    self.__queue.task_done()
                    if m[0] == 'PUBLISH':
                        ntopic = self.__deftopic
                        if m[1] is not None:  # topic is set
                            ntopic = m[1]
                        nqos = m[3]
                        if nqos is None:
                            nqos = self.__qos
                        if ntopic:
                            msg = None
                            if m[2] is not None:
                                msg = m[2].encode('utf-8')

                            with self.__publishLock:
                                mi = self.__client.publish(
                                    ntopic, msg, nqos, m[4])

                                # QoS == 0 when not connected are discarded
                                if nqos != 0 or mi.rc == 0:
                                    self.__unsent.add(mi.mid)

                            # Wait for publish if timeout provided
                            if m[5] is not None and m[5] > 0:
                                if mi.rc == 0:
                                    mi.wait_for_publish(m[5])
                                else:
                                    if nqos == 0:
                                        _log.warning(
                                            'Publish to %r dropped: %r,%r',
                                            ntopic, mi.rc, mi.mid)
                    elif m[0] == 'SUBSCRIBE':
                        _log.debug('Subscribe topic: %r q=%r', m[1], m[2])
                        nqos = m[2]
                        if nqos is None:
                            nqos = self.__qos
                        self.__client.subscribe(m[1], nqos)
                    elif m[0] == 'UNSUBSCRIBE':
                        _log.debug('Un-subscribe topic: %r', m[1])
                        self.__client.unsubscribe(m[1])
                    elif m[0] == 'RECONNECT':
                        self.__connect_pending = False
                        self.__doreconnect = True
                    elif m[0] == 'EXIT':
                        _log.debug('Request to close: %r', m[1])
                        self.__running = False
            except queue.Empty:
                pass
            except Exception as e:
                _log.error('%s: %s', e.__class__.__name__, e)
                self.__connect_pending = False
                self.__doreconnect = False
        if self.__connected:
            self.__client.disconnect()
            self.__client.loop_stop()
        _log.info('Exiting')
